from unittest.mock import patch

import numpy as np
import pytest
import torch

from ml4gw.dataloading.in_memory_dataset import InMemoryDataset


@pytest.fixture(params=[1, 10])
def num_kernels(request):
    return request.param


@pytest.fixture
def sample_rate():
    return 128


@pytest.fixture(params=[8])
def kernel_size(request):
    return request.param


@pytest.fixture(params=[1, 2])
def stride(request):
    return request.param


@pytest.fixture(params=[0, 1, -1])
def extra(request):
    return request.param


@pytest.fixture(params=[1, 4])
def batch_size(request):
    return request.param


@pytest.fixture
def X(num_kernels, stride, kernel_size, extra):
    num_samples = (num_kernels - 1) * stride + kernel_size
    if extra == -1:
        num_samples += stride - 1
    elif extra < stride:
        num_samples += extra

    x = torch.arange(num_samples)
    return torch.stack([i * num_samples + x for i in range(3)])


@pytest.fixture(params=[True, False])
def use_y(request):
    return request.param


@pytest.fixture
def Xy(X, use_y):
    y = X[-1] + 1 if use_y else None
    return X, y


@pytest.fixture
def validate_shape(kernel_size, batch_size, num_kernels):
    def f(i, x, length, y=None):
        if batch_size == 1 or (i + 1) < length:
            expected_batch = batch_size
        elif batch_size > num_kernels:
            expected_batch = num_kernels
        else:
            expected_batch = 2

        assert x.shape == (expected_batch, 3, kernel_size)
        if y is not None:
            assert y.shape == (expected_batch, kernel_size)

    return f


def test_in_memory_dataset_coincident_deterministic(
    Xy, num_kernels, kernel_size, stride, batch_size, validate_shape
):
    X, y = Xy
    if y is not None:
        # ensure that y has to be same length as x in time
        with pytest.raises(ValueError) as exc:
            dataset = InMemoryDataset(
                X,
                kernel_size,
                y=y[:-1],
                batch_size=batch_size,
                stride=stride,
                coincident=True,
                shuffle=False,
            )
        assert str(exc.value).startswith("Target timeseries")

    dataset = InMemoryDataset(
        X,
        kernel_size,
        y=y,
        batch_size=batch_size,
        stride=stride,
        coincident=True,
        shuffle=False,
    )
    # validate some attributes on the dataset
    assert dataset.num_kernels == num_kernels
    if y is None:
        assert dataset.y is None

    if batch_size == 1 or batch_size > num_kernels:
        assert len(dataset) == num_kernels
    else:
        assert len(dataset) == 3

    # now iterate through and make sure all the
    # produced arrays match our expectations
    num_samples = X.shape[-1]
    for i, x in enumerate(dataset):
        if y is not None:
            x, y_ = x
            y_ = y_.cpu().numpy()
        else:
            y_ = None

        validate_shape(i, x, len(dataset), y_)

        for j, sample in enumerate(x.cpu().numpy()):
            for k, channel in enumerate(sample):
                start = i * batch_size * stride + j * stride + k * num_samples
                stop = start + kernel_size
                expected = np.arange(start, stop).astype("float32")
                assert (channel == expected).all()

            # y should be equal to the last channel plus 1
            if y is not None:
                assert (y_[j] == (expected + 1)).all()

    assert (i + 1) == len(dataset)


@pytest.fixture(params=[1, 4])
def batches_per_epoch(request):
    return request.param


def test_in_memory_dataset_non_coincident_deterministic(
    X,
    num_kernels,
    kernel_size,
    stride,
    batch_size,
    validate_shape,
    batches_per_epoch,
):
    # ensure that we can't pass a target to a non-coincident dataset
    with pytest.raises(ValueError) as exc:
        dataset = InMemoryDataset(
            X,
            kernel_size,
            y=X[-1],
            batch_size=batch_size,
            batches_per_epoch=batches_per_epoch,
            stride=stride,
            coincident=False,
            shuffle=False,
        )
    assert str(exc.value).startswith("Can't sample target array")

    # ensure that we have to pass batches_per_epoch
    with pytest.raises(ValueError) as exc:
        dataset = InMemoryDataset(
            X,
            kernel_size,
            batch_size=batch_size,
            stride=stride,
            coincident=False,
            shuffle=False,
        )
    assert str(exc.value).startswith("Must specify number of batches")

    if num_kernels < (batch_size * batches_per_epoch):
        with pytest.raises(ValueError):
            dataset = InMemoryDataset(
                X,
                kernel_size,
                batch_size=batch_size,
                stride=stride,
                coincident=False,
                shuffle=False,
                batches_per_epoch=batches_per_epoch,
            )
        return

    # create, validate attributes, validate returned values
    dataset = InMemoryDataset(
        X,
        kernel_size,
        batch_size=batch_size,
        stride=stride,
        coincident=False,
        shuffle=False,
        batches_per_epoch=batches_per_epoch,
    )
    assert dataset.num_kernels == num_kernels
    assert dataset.y is None
    assert len(dataset) == batches_per_epoch

    num_samples = X.shape[-1]
    for i, x in enumerate(dataset):
        assert x.shape == (batch_size, 3, kernel_size)

        for j, sample in enumerate(x.cpu().numpy()):
            sample_idx = i * batch_size + j
            for k, channel in enumerate(sample):
                sample_step = sample_idx // (num_kernels ** (2 - k))
                start = k * num_samples + sample_step * stride
                stop = start + kernel_size
                expected = np.arange(start, stop).astype("float32")
                assert (channel == expected).all()

    assert (i + 1) == len(dataset)


def test_in_memory_dataset_coincident_stochastic(
    Xy, num_kernels, kernel_size, stride, batch_size, validate_shape
):
    # don't do some of the checks on errors or attributes
    # since the shuffle flag doesn't affect any of these;
    X, y = Xy

    # patch return value of init_indices,
    # (in this case, will call randperm) so that
    # we know exactly what the iteration order will be
    idx = torch.randperm(num_kernels)
    with patch.object(InMemoryDataset, "init_indices", return_value=idx):
        dataset = InMemoryDataset(
            X,
            kernel_size,
            y=y,
            batch_size=batch_size,
            stride=stride,
            coincident=True,
            shuffle=True,
        )

        data_it = iter(dataset)
        idx = idx.cpu().numpy()

        # now iterate through the initialized iterator and
        # use the sampled indices to find where in the
        # arange array we ought to be
        num_samples = X.shape[-1]
        for i in range(len(dataset)):
            x = next(data_it)
            if y is not None:
                x, y_ = x
                y_ = y_.cpu().numpy()
            else:
                y_ = None

            validate_shape(i, x, len(dataset), y_)
            for j, sample in enumerate(x.cpu().numpy()):
                for k, channel in enumerate(sample):
                    sample_idx = i * batch_size + j
                    start_idx = idx[sample_idx]
                    start = start_idx * stride + k * num_samples
                    stop = start + kernel_size
                    expected = np.arange(start, stop).astype("float32")
                    assert (channel == expected).all(), (i, j, k)

                if y is not None:
                    assert (y_[j] == (expected + 1)).all()

        with pytest.raises(StopIteration):
            next(data_it)


def test_in_memory_dataset_non_coincident_stochastic(
    X,
    num_kernels,
    kernel_size,
    stride,
    batch_size,
    validate_shape,
    batches_per_epoch,
):
    # skip checks besides this one since we checked
    # errors and attributes earlier
    if num_kernels < (batch_size * batches_per_epoch):
        with pytest.raises(ValueError):
            dataset = InMemoryDataset(
                X,
                kernel_size,
                batch_size=batch_size,
                stride=stride,
                coincident=False,
                shuffle=False,
                batches_per_epoch=batches_per_epoch,
            )
        return

    # lets make sure we're correctly calling
    # torch.randint function for
    # coincident=True, shuffle=False use case
    samples = batches_per_epoch * batch_size
    idx = torch.randint(num_kernels, size=(samples, 3))
    dataset = InMemoryDataset(
        X,
        kernel_size,
        batch_size=batch_size,
        stride=stride,
        coincident=False,
        shuffle=True,
        batches_per_epoch=batches_per_epoch,
    )
    with patch("torch.randint", return_value=idx) as mock:
        dataset.init_indices()

    mock.assert_called_once_with(
        num_kernels, size=(samples, 3), device=dataset.X.device
    )

    # Now test that the iteration order is as expected
    # with the patched indices
    with patch.object(InMemoryDataset, "init_indices", return_value=idx):
        dataset = InMemoryDataset(
            X,
            kernel_size,
            batch_size=batch_size,
            stride=stride,
            coincident=False,
            shuffle=True,
            batches_per_epoch=batches_per_epoch,
        )

        idx = idx.cpu().numpy()
        data_it = iter(dataset)
        # use the sampled indices to make sure the iteration
        # order matches our expectations
        num_samples = X.shape[-1]
        for i in range(len(dataset)):
            x = next(data_it)
            assert x.shape == (batch_size, 3, kernel_size)

            for j, sample in enumerate(x.cpu().numpy()):
                for k, channel in enumerate(sample):
                    sample_idx = i * batch_size + j
                    start_idx = idx[sample_idx, k]
                    start = k * num_samples + start_idx * stride
                    stop = start + kernel_size
                    expected = np.arange(start, stop).astype("float32")
                    assert (channel == expected).all()

    with pytest.raises(StopIteration):
        next(data_it)



from unittest.mock import patch

import pytest
import torch

from ml4gw.spectral import spectral_density
from ml4gw.transforms import FixedWhiten, Whiten


class WhitenModuleTest:
    sample_rate = 8192
    psd_length = 128
    whiten_length = 64
    num_channels = 5

    mean = 2
    std = 5

    @pytest.fixture
    def background(self):
        background_size = int(self.psd_length * self.sample_rate)
        x = [torch.randn(background_size) for _ in range(self.num_channels)]
        return [self.mean + self.std * i for i in x]

    @pytest.fixture
    def X(self):
        size = int(self.whiten_length * self.sample_rate)
        X = torch.randn(8, self.num_channels, size)
        return self.mean + self.std * X

    @pytest.fixture(params=[None, 32])
    def highpass(self, request):
        return request.param

    @pytest.fixture(params=[None, 512])
    def lowpass(self, request):
        return request.param

    def get_psds(self, background, fftlength):
        nperseg = int(fftlength * self.sample_rate)
        window = torch.hann_window(nperseg)
        psds = []
        for bg in background:
            psd = spectral_density(
                bg,
                nperseg,
                nperseg // 2,
                window,
                scale=1 / (self.sample_rate * (window**2).sum()),
            )
            psds.append(psd)
        return psds


class TestWhiten(WhitenModuleTest):
    fduration = 1

    @pytest.fixture
    def transform(self, highpass, lowpass):
        return Whiten(self.fduration, self.sample_rate, highpass, lowpass)

    def test_init(self, transform):
        assert transform.window.size(0) == 8192

    def test_forward(
        self, transform, X, background, highpass, lowpass, validate_whitened
    ):
        background = self.get_psds(background, 2)
        background = torch.stack(background)
        whitened = transform(X, background)
        assert whitened.shape == (
            8,
            5,
            X.size(-1) - self.fduration * self.sample_rate,
        )
        validate_whitened(
            whitened,
            highpass,
            lowpass,
            self.sample_rate,
            1 / self.whiten_length,
        )


class TestFixedWhiten(WhitenModuleTest):
    def get_transform(self):
        return FixedWhiten(
            self.num_channels, self.whiten_length, self.sample_rate
        )

    @pytest.fixture
    def transform(self):
        return self.get_transform()

    def test_init(self, transform, X):
        # ensure parameters have been initialized
        # with the right shapes and to 0 valuess
        num_freqs = (8192 * 64) // 2 + 1
        assert transform.psd.shape == (self.num_channels, num_freqs)
        assert (transform.psd == 0).all().item()

        assert transform.fduration.shape == (1,)
        assert (transform.fduration == 0).item()

        # ensure that trying to call forward
        # before fitting raises an error
        assert not transform.built
        with pytest.raises(ValueError) as exc:
            transform(X)
        assert str(exc.value).startswith("Must fit parameters")

    def test_fit_time_domain(
        self, transform, background, X, highpass, lowpass, validate_whitened
    ):
        # ensure that calling with the wrong number
        # of background channels raises an error
        with pytest.raises(ValueError) as exc:
            transform.fit(2, *background[:-1], fftlength=4)
        assert str(exc.value).startswith("Expected to fit whitening")

        # fit to the background and ensure that the
        # parameter values have been set to nonzero values
        transform.fit(
            2, *background, fftlength=2, highpass=highpass, lowpass=lowpass
        )
        assert (transform.psd != 0).all().item()
        assert (transform.fduration == 2).item()

        # now whiten a dummy tensor and make sure
        # that the values come out as expected
        whitened = transform(X)
        validate_whitened(
            whitened,
            highpass,
            lowpass,
            self.sample_rate,
            1 / self.whiten_length,
        )

    def test_fit_freq_domain(
        self, transform, background, X, highpass, lowpass, validate_whitened
    ):
        # first check if fftlength == self.whiten_length,
        # the fit psd should match the psds used to fit
        # almost exactly if we don't use inverse spectrum
        # truncation
        psds = self.get_psds(background, self.whiten_length)
        with patch(
            "ml4gw.transforms.whitening.spectral.truncate_inverse_power_spectrum",
            new=lambda x, _, __, ___, ____: x,
        ):
            transform.fit(2, *psds)
        assert (transform.fduration == 2).item()
        for i, psd in enumerate(psds):
            torch.testing.assert_close(
                psd, transform.psd[i], rtol=1e-6, atol=0.0, check_dtype=False
            )

        # now do a fit with a more realistic
        # fftlength, fduration, highpass, and lowpass
        psds = self.get_psds(background, 2)
        transform.fit(2, *psds, highpass=highpass, lowpass=lowpass)
        assert (transform.psd != 0).all().item()
        assert (transform.fduration == 2).item()

        # now whiten a dummy tensor and make sure
        # that the values come out as expected
        whitened = transform(X)
        validate_whitened(
            whitened,
            highpass,
            lowpass,
            self.sample_rate,
            1 / self.whiten_length,
        )

    def test_io(self, transform, background, tmp_path):
        transform.fit(2, *background, fftlength=4)
        assert transform.built
        tmp_path.mkdir(exist_ok=True, parents=True)
        torch.save(transform.state_dict(), tmp_path / "whiten.pt")

        fresh = self.get_transform()
        assert not fresh.built
        assert (fresh.fduration == 0).item()
        state_dict = torch.load(tmp_path / "whiten.pt")
        fresh.load_state_dict(state_dict)
        assert fresh.built

        assert (fresh.fduration == 2).item()
        assert (fresh.psd == transform.psd).all().item()



from unittest.mock import patch

import pytest
import torch

from ml4gw.transforms.waveforms import WaveformProjector, WaveformSampler


@pytest.fixture(params=[100, 1000])
def num_waveforms(request):
    return request.param


@pytest.fixture(params=[1024, 2048])
def sample_rate(request):
    return request.param


@pytest.fixture(params=[["H1"], ["H1", "L1"]])
def ifos(request):
    return request.param


def test_waveform_sampler(num_waveforms):
    # test that instantiating with different numbers of waveforms
    # for each polarization raises an error
    wrong = {str(i): torch.randn(i + 2, 1024) for i in range(2)}
    with pytest.raises(ValueError) as exc:
        WaveformSampler(**wrong)
    assert str(exc.value).startswith("Polarization")

    # instantiate waveform sampler
    waveforms = {
        i: torch.randn(num_waveforms, 1024) for i in ["plus", "cross"]
    }
    waveform_sampler = WaveformSampler(**waveforms)
    assert waveform_sampler.num_waveforms == num_waveforms
    assert waveform_sampler.parameters is None

    # test that requesting more waveforms than are available raises an error
    with pytest.raises(ValueError) as exc:
        waveform_sampler(num_waveforms + 1)
    assert str(exc.value).startswith("Requested")

    # test that sampling produces expected output
    samples = waveform_sampler(num_waveforms)
    for key in ["plus", "cross"]:
        assert samples[key].shape == (num_waveforms, 1024)

    # test that intrinsic parameters are returned if provided
    num_parameters = 5
    intrinsic_parameters = torch.column_stack(
        [torch.arange(0, num_waveforms, 1) * i for i in range(num_parameters)]
    )
    waveform_sampler = WaveformSampler(
        **waveforms, parameters=intrinsic_parameters
    )
    randint_patch = patch("torch.randint", return_value=torch.arange(10))
    with randint_patch:
        samples, params = waveform_sampler(10)

    assert len(samples["plus"]) == len(samples["cross"]) == len(params) == 10
    assert (params == (intrinsic_parameters[:10, :])).all()

    # test that passing in the wrong number of parameters raises an error
    wrong = torch.column_stack(
        [
            torch.arange(0, num_waveforms + 1, 1) * i
            for i in range(num_parameters)
        ]
    )
    with pytest.raises(ValueError) as exc:
        waveform_sampler = WaveformSampler(parameters=wrong, **waveforms)
    assert str(exc.value).startswith("Waveform parameters")


@pytest.fixture(autouse=True)
def compute_observed_strain(ifos):
    def f(*args, **kwargs):
        waveform = kwargs["plus"] + kwargs["cross"]
        return torch.stack([waveform + i for i in range(len(ifos))], axis=1)

    with patch("ml4gw.gw.compute_observed_strain", new=f):
        yield


def test_waveform_projector(ifos, sample_rate):
    projector = WaveformProjector(ifos, sample_rate)

    waveforms = {i: torch.randn(100, 1024) for i in ["plus", "cross"]}
    dec = torch.randn(len(waveforms["plus"]))
    psi = torch.randn(len(waveforms["plus"]))
    phi = torch.randn(len(waveforms["plus"]))

    waveform = waveforms["plus"] + waveforms["cross"]
    summed = torch.stack([waveform + i for i in range(len(ifos))], axis=1)

    projected = projector(dec, psi, phi, **waveforms)

    assert projected.shape == (100, len(ifos), 1024)
    assert (projected == summed).all()



import pytest

from ml4gw.transforms import SnrRescaler


@pytest.fixture(params=[128, 512])
def sample_rate(request):
    return request.param


@pytest.fixture(params=[1, 2])
def waveform_duration(request):
    return request.param


@pytest.fixture(params=[["H1"], ["H1", "L1"]])
def ifos(request):
    return request.param


class TestSnrRescaler:
    @pytest.fixture(params=[1, 2])
    def num_channels(self, request):
        return request.param

    @pytest.fixture(params=[128, 512])
    def sample_rate(self, request):
        return request.param

    @pytest.fixture(params=[1, 2])
    def waveform_duration(self, request):
        return request.param

    @pytest.fixture(params=[None, 32])
    def highpass(self, request):
        return request.param

    @pytest.fixture(params=[None, 64])
    def lowpass(self, request):
        return request.param

    @pytest.fixture
    def transform(
        self, num_channels, sample_rate, waveform_duration, highpass, lowpass
    ):
        return SnrRescaler(
            num_channels, sample_rate, waveform_duration, highpass, lowpass
        )

    def test_init(
        self,
        transform,
        num_channels,
        sample_rate,
        waveform_duration,
        highpass,
        lowpass,
    ):
        assert not transform.built

        num_freqs = waveform_duration * sample_rate // 2 + 1
        shape = (num_channels, num_freqs)
        assert transform.background.shape == shape
        assert (transform.background == 0).all().item()

        if highpass is not None and lowpass is not None:
            start = int(highpass * waveform_duration)
            stop = int(lowpass * waveform_duration)
            assert len(transform._buffers) == 3
            assert not (transform.highpass_mask[:start].any().item())
            assert not (transform.lowpass_mask[stop:].any().item())
            assert transform.lowpass_mask[start:stop].all().item()
        elif highpass is not None:
            idx = int(highpass * waveform_duration)
            assert len(transform._buffers) == 2
            assert not (transform.highpass_mask[:idx]).any().item()
            assert (transform.highpass_mask[idx:]).all().item()
        elif lowpass is not None:
            idx = int(lowpass * waveform_duration)
            assert len(transform._buffers) == 2
            assert not (transform.lowpass_mask[idx:]).any().item()
            assert (transform.lowpass_mask[:idx]).all().item()
        else:
            assert len(transform._buffers) == 1
            assert transform.highpass_mask is None
            assert transform.lowpass_mask is None



import random

import numpy as np
import pytest
import torch
from scipy.special import erfinv
from torch.distributions import Uniform


# If a fixture is doing anything random,
# it should take this function as an argument
@pytest.fixture(autouse=True)
def seed_everything():
    seed = 101589
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


@pytest.fixture
def compare_against_numpy():
    """
    idea here is that if relative error is
    distributed as a zero mean gaussian with
    variance sigma, pick a tolerance such that
    all values will fall into spec prob fraction
    of the time
    """

    def compare(value, expected, num_bad: int = 0):
        sigma = 0.01
        prob = 0.99999
        N = np.prod(expected.shape)
        tol = sigma * erfinv(prob ** (1 / N)) * 2**0.5

        isclose = np.isclose(value, expected, rtol=tol)

        # at most one point can differ by more than tolerance
        # this happens occasionally and typically for very low values

        # TODO: eventually we should track down
        # and address the underlying cause
        assert isclose.sum() - np.prod(isclose.shape) <= num_bad

    return compare


@pytest.fixture
def validate_whitened():
    def validate(whitened, highpass, lowpass, sample_rate, df):
        # make sure we have 0 mean unit variance
        means = whitened.mean(axis=-1)
        target = torch.zeros_like(means)
        torch.testing.assert_close(means, target, rtol=0, atol=0.02)

        stds = whitened.std(axis=-1)
        target = torch.ones_like(stds)

        # if we're highpassing or lowpassing, then we
        # shouldn't expect the standard deviation to be
        # one because we're subtracting some power, so
        # remove roughly the expected power contributed
        # by the highpassed frequencies from the target.
        if highpass is not None:
            nyquist = sample_rate / 2
            target *= (1 - highpass / nyquist) ** 0.5
        if lowpass is not None:
            nyquist = sample_rate / 2
            target *= (lowpass / nyquist) ** 0.5

        # TODO: most statistically accurate test would be
        # to ensure that variances of the whitened data
        # along the time dimension are distributed like
        # a chi-squared with degrees of freedom equal to
        # the number of samples along time dimension, but
        # there's extra variance to account for in the
        # PSD as well that throws this off. There should be
        # a way to account for all of these sources of noise
        # in the tolerance, but for now we'll just adopt the
        # tolerance that gwpy uses in its tests
        torch.testing.assert_close(stds, target, rtol=0.04, atol=0.0)

        # check that frequencies up to close to the highpass/lowpass
        # frequencies have near 0 power.
        # TODO: the tolerance will need to increase
        # the closer to the cutoff frequency we get,
        # so what's a better way to check this
        if highpass is not None:
            fft = torch.fft.rfft(whitened, norm="ortho").abs()
            idx = int(0.8 * highpass / df)
            passed = fft[:, :, :idx]
            target = torch.zeros_like(passed)
            torch.testing.assert_close(passed, target, rtol=0, atol=0.07)
        if lowpass is not None:
            fft = torch.fft.rfft(whitened, norm="ortho").abs()
            idx = int(1.2 * lowpass / df)
            passed = fft[:, :, idx:]
            target = torch.zeros_like(passed)
            torch.testing.assert_close(passed, target, rtol=0, atol=0.07)

    return validate


# A num_samples fixture should be defined for any
# test that wants to use these fixtures


@pytest.fixture(params=[256, 1024, 2048])
def sample_rate(request):
    return request.param


@pytest.fixture()
def chirp_mass(num_samples, seed_everything):
    dist = Uniform(5, 100)
    return dist.sample((num_samples,))


@pytest.fixture()
def mass_ratio(num_samples, seed_everything):
    dist = Uniform(0.125, 0.99)
    return dist.sample((num_samples,))


@pytest.fixture()
def a_1(num_samples, seed_everything):
    dist = Uniform(0, 0.90)
    return dist.sample((num_samples,))


@pytest.fixture()
def a_2(num_samples, seed_everything):
    dist = Uniform(0, 0.90)
    return dist.sample((num_samples,))


@pytest.fixture()
def tilt_1(num_samples, seed_everything):
    dist = Uniform(0, torch.pi)
    return dist.sample((num_samples,))


@pytest.fixture()
def tilt_2(num_samples, seed_everything):
    dist = Uniform(0, torch.pi)
    return dist.sample((num_samples,))


@pytest.fixture()
def phi_12(num_samples, seed_everything):
    dist = Uniform(0, 2 * torch.pi)
    return dist.sample((num_samples,))


@pytest.fixture()
def phi_jl(num_samples, seed_everything):
    dist = Uniform(0, 2 * torch.pi)
    return dist.sample((num_samples,))


@pytest.fixture()
def distance(num_samples, seed_everything):
    dist = Uniform(100, 3000)
    return dist.sample((num_samples,))


@pytest.fixture()
def distance_far(num_samples, seed_everything):
    dist = Uniform(500, 3000)
    return dist.sample((num_samples,))


@pytest.fixture()
def distance_close(num_samples, seed_everything):
    dist = Uniform(100, 500)
    return dist.sample((num_samples,))


@pytest.fixture()
def theta_jn(num_samples, seed_everything):
    dist = Uniform(0, torch.pi)
    return dist.sample((num_samples,))


@pytest.fixture()
def phase(num_samples, seed_everything):
    dist = Uniform(0, 2 * torch.pi)
    return dist.sample((num_samples,))


@pytest.fixture()
def chi1(num_samples, seed_everything):
    dist = Uniform(-0.999, 0.999)
    return dist.sample((num_samples,))


@pytest.fixture()
def chi2(num_samples, seed_everything):
    dist = Uniform(-0.999, 0.999)
    return dist.sample((num_samples,))



import pytest
import torch

from ml4gw.transforms.scaler import ChannelWiseScaler


def test_scaler_1d():
    scaler = ChannelWiseScaler()
    assert len(list(scaler.buffers())) == 2
    assert scaler.mean.shape == (1,)
    assert scaler.std.shape == (1,)

    # test fitting
    background = torch.arange(1, 11).type(torch.float32)
    scaler.fit(background)

    mean, std = 5.5, (99 / 12) ** 0.5
    assert scaler.mean.item() == pytest.approx(mean)
    assert scaler.std.item() == pytest.approx(std)

    # test fit shape checks
    with pytest.raises(AssertionError):
        scaler.fit(background[None])
    with pytest.raises(ValueError):
        scaler.fit(background[None, None])

    # test forward
    x = torch.arange(20).reshape(2, 10).type(torch.float32) + 1
    y = scaler(x)
    for i in range(2):
        expected = i * 10 + torch.arange(1, 11)
        expected = (expected - mean) / std
        assert (y[i] == expected).all().item()

    # now reverse it
    y = scaler(y, reverse=True)
    assert torch.isclose(x, y, rtol=1e-6).all().item()


def test_scaler_regularization():
    scaler_non_reg = ChannelWiseScaler(num_channels=2)
    scaler_reg = ChannelWiseScaler(num_channels=2)
    X = torch.ones(2, 100).type(torch.float32)
    scaler_non_reg.fit(X)
    scaler_reg.fit(X, std_reg=1e-8)
    assert torch.all(scaler_non_reg.std == 0)
    assert torch.all(scaler_reg.std > 0)


def test_scaler_2d():
    num_channels = 4
    scaler = ChannelWiseScaler(num_channels)
    assert len(list(scaler.buffers())) == 2
    assert scaler.mean.shape == (num_channels, 1)
    assert scaler.std.shape == (num_channels, 1)

    # test fitting
    background = (
        torch.arange(num_channels * 10)
        .reshape(num_channels, 10)
        .type(torch.float32)
        + 1
    )
    scaler.fit(background)

    std = (99 / 12) ** 0.5
    for i in range(num_channels):
        mean = ((2 * i + 1) * 10 + 1) / 2
        assert scaler.mean[i, 0] == pytest.approx(mean)
        assert scaler.std[i, 0] == pytest.approx(std)

    # test fit shape checks
    with pytest.raises(AssertionError):
        scaler.fit(background[0])
    with pytest.raises(ValueError):
        scaler.fit(background[None, None])

    # test forward with 2D
    x = torch.arange(num_channels * 10).type(torch.float32)
    x = x.reshape(num_channels, 10).type(torch.float32) + 1
    y = scaler(x)
    expected = (torch.arange(1, 11) - 5.5) / std
    assert torch.isclose(y, expected, rtol=1e-6).all().item()

    # now reverse it
    y = scaler(y, reverse=True)
    assert torch.isclose(x, y, rtol=1e-6).all().item()

    # test forward with 3D
    x = torch.stack([x, x + 1])
    y = scaler(x)
    expected = torch.stack([expected, expected + 1 / std])
    assert torch.isclose(y.transpose(1, 0), expected, rtol=1e-6).all().item()

    # now reverse
    y = scaler(y, reverse=True)
    assert torch.isclose(x, y, rtol=1e-6).all().item()


def test_scaler_save_and_load(tmp_path):
    scaler = ChannelWiseScaler()
    background = torch.arange(1, 11).type(torch.float32)

    scaler.fit(background)
    assert scaler.built
    mean, std = 5.5, (99 / 12) ** 0.5

    tmp_path.mkdir(parents=True, exist_ok=True)
    torch.save(scaler.state_dict(), tmp_path / "scaler.pt")

    scaler = ChannelWiseScaler()
    assert not scaler.built
    assert (scaler.mean == 0).all().item()
    assert (scaler.std == 1).all().item()

    scaler.load_state_dict(torch.load(tmp_path / "scaler.pt"))
    assert scaler.built
    assert scaler.mean == mean
    assert scaler.std == std

    scaler = ChannelWiseScaler(2)
    with pytest.raises(RuntimeError):
        scaler.load_state_dict(torch.load(tmp_path / "scaler.pt"))



import math

import numpy as np
import pytest
import torch
from scipy import optimize

from ml4gw import distributions

# TODO: for all tests, how to validate that
# distribution has the expected shape?


def test_log_uniform():
    sampler = distributions.LogUniform(math.e, math.e**2)
    samples = sampler.sample((10,))
    assert len(samples) == 10
    assert ((torch.e <= samples) & (torch.e**2 <= 100)).all()

    # check that the mean is roughly correct
    # (within three standard deviations)
    samples = sampler.sample((100000,))
    log_samples = np.log(samples)

    mean = log_samples.mean().item()
    variance = 4 / 12
    sample_variance = variance / 10000
    sample_std = sample_variance**0.5
    assert abs(mean - 1.5) < (3 * sample_std)


def test_cosine():
    sampler = distributions.Cosine()
    samples = sampler.sample((10,))
    assert len(samples) == 10
    assert ((-math.pi / 2 <= samples) & (samples <= math.pi / 2)).all()

    sampler = distributions.Cosine(-3, 5)
    samples = sampler.sample((100,))
    assert len(samples) == 100
    assert ((-3 <= samples) & (samples <= 5)).all()

    assert torch.all(sampler.log_prob(torch.tensor([-4, 6])) == float("-inf"))


def test_power_law():
    """Test PowerLaw distribution"""
    ref_snr = 8
    sampler = distributions.PowerLaw(ref_snr, float("inf"), index=-4)
    samples = sampler.sample((100000,)).numpy()
    # check x^-4 behavior
    counts, ebins = np.histogram(samples, bins=1000)
    bins = ebins[1:] + ebins[:-1]
    bins *= 0.5

    def foo(x, a, b):
        return a * x**b

    popt, _ = optimize.curve_fit(foo, bins, counts)
    # popt[1] is the index
    assert popt[1] == pytest.approx(-4, rel=1e-1)

    min_dist = 10
    max_dist = 1000
    uniform_in_volume = distributions.PowerLaw(min_dist, max_dist, index=2)
    samples = uniform_in_volume.sample((100000,)).numpy()
    # check d^2 behavior
    counts, ebins = np.histogram(samples, bins=1000)
    bins = ebins[1:] + ebins[:-1]
    bins *= 0.5

    popt, _ = optimize.curve_fit(foo, bins, counts)
    # popt[1] is the index
    assert popt[1] == pytest.approx(2, rel=1e-1)

    # test 1/x distribution
    inverse_in_distance = distributions.PowerLaw(min_dist, max_dist, index=-1)
    samples = inverse_in_distance.sample((100000,)).numpy()
    counts, ebins = np.histogram(samples, bins=1000)
    bins = ebins[1:] + ebins[:-1]
    bins *= 0.5
    popt, _ = optimize.curve_fit(foo, bins, counts)

    assert popt[1] == pytest.approx(-1, rel=1e-1)


def test_delta_function():
    sampler = distributions.DeltaFunction(peak=20)
    samples = sampler.sample((10,))
    assert (samples == 20).all()



from io import BytesIO

import numpy as np
import pytest
import torch
from scipy import signal

from ml4gw.transforms.spectral import SpectralDensity

TOL = 1e-7


@pytest.fixture(params=[1, 4, 8])
def length(request):
    return request.param


@pytest.fixture(params=[1024, 4096])
def sample_rate(request):
    return request.param


@pytest.mark.parametrize(
    "overlap,fftlength",
    [[None, 2], [0.5, 2], [2, 2], [None, 4], [2, 4], [4, 4]],
)
def test_init(overlap, fftlength, sample_rate):
    if overlap is not None and overlap >= fftlength:
        with pytest.raises(ValueError):
            transform = SpectralDensity(sample_rate, fftlength, overlap)
        return
    else:
        transform = SpectralDensity(sample_rate, fftlength, overlap)

    if overlap is None:
        expected_stride = int(fftlength * sample_rate // 2)
    else:
        expected_stride = int(fftlength * sample_rate) - int(
            overlap * sample_rate
        )
    assert transform.nstride == expected_stride
    assert len(list(transform.buffers())) == 2

    # test read/write
    weights_io = BytesIO()
    torch.save(transform.state_dict(), weights_io)

    weights_io.seek(0)
    transform.load_state_dict(torch.load(weights_io))


@pytest.fixture(params=[0.5, 2, 4])
def fftlength(request):
    return request.param


@pytest.fixture(params=[None, 0.1, 0.5, 1])
def overlap(request):
    return request.param


@pytest.fixture(params=[None, 55, [25, 55]])
def freq_low(request):
    return request.param


@pytest.fixture(params=[None, 65, [35, 65]])
def freq_high(request):
    return request.param


@pytest.fixture(params=[True, False])
def fast(request):
    return request.param


@pytest.fixture(params=["mean", "median"])
def average(request):
    return request.param


@pytest.fixture(params=[1, 2, 3])
def ndim(request):
    return request.param


@pytest.fixture(params=["hann"])
def window(request, fftlength, sample_rate):
    return signal.get_window(request.param, int(fftlength * sample_rate))


def test_spectral_density(
    length, sample_rate, fftlength, overlap, fast, average, ndim, window
):
    batch_size = 8
    num_channels = 5
    if overlap is not None and overlap >= fftlength:
        return

    transform = SpectralDensity(
        sample_rate,
        fftlength,
        overlap,
        average=average,
        window=torch.Tensor(window),
        fast=fast,
    )

    shape = [int(length * sample_rate)]
    if ndim > 1:
        shape.insert(0, num_channels)
    if ndim > 2:
        shape.insert(0, batch_size)
    x = np.random.randn(*shape)

    # make sure we catch if the fftlength is too long for the data
    if fftlength > length:
        with pytest.raises(ValueError) as exc_info:
            transform(torch.Tensor(x))
        assert str(exc_info.value).startswith("Number of samples")
        return

    # perform the transform and confirm the shape is correct
    torch_result = transform(torch.Tensor(x)).numpy()
    num_freq_bins = int(fftlength * sample_rate) // 2 + 1
    shape[-1] = num_freq_bins
    assert torch_result.shape == tuple(shape)

    # now verify against the result from scipy
    _, scipy_result = signal.welch(
        x,
        fs=sample_rate,
        nperseg=transform.nperseg,
        noverlap=transform.nperseg - transform.nstride,
        window=window,
        average=average,
    )

    # if we're using the fast implementation, only guarantee
    # that components higher than the first two are correct
    if fast:
        torch_result = torch_result[..., 2:]
        scipy_result = scipy_result[..., 2:]
    assert np.isclose(torch_result, scipy_result, rtol=TOL).all()

    # test error raised if window incorrect shape
    with pytest.raises(ValueError) as exc_info:
        window = torch.hann_window(1)
        transform = SpectralDensity(
            sample_rate,
            fftlength,
            overlap,
            average=average,
            window=window,
            fast=fast,
        )


@pytest.fixture(params=[0, 1])
def y_ndim(request):
    return request.param


def _shape_checks(ndim, y_ndim, x, y, transform):
    # verify that time dimensions must match
    with pytest.raises(ValueError) as exc_info:
        transform(x, y[..., :-1])
    assert str(exc_info.value).startswith("Time dimensions")

    # verify that y can't have more dims than x
    if y_ndim == 0:
        with pytest.raises(ValueError) as exc_info:
            transform(x, y[None])
        assert str(exc_info.value).startswith("Can't compute")

        if ndim == 1:
            assert "1D" in str(exc_info.value)

    # verify that if x is greater than 1D and y has
    # the same dimensionality, their shapes must
    # fully match
    if ndim > 1 and y_ndim == 0:
        with pytest.raises(ValueError) as exc_info:
            transform(x, y[:-1])
        assert str(exc_info.value).startswith("If x and y tensors")

    # verify for 3D x's that 2D y's must have the same batch
    # dimension, and that y cannot be 1D
    if ndim == 3 and y_ndim == 1:
        with pytest.raises(ValueError) as exc_info:
            transform(x, y[:-1])
        assert str(exc_info.value).startswith("If x is a 3D tensor")

        with pytest.raises(ValueError) as exc_info:
            transform(x, y[0])
        assert str(exc_info.value).startswith("Can't compute cross")


def test_transform_with_csd(
    y_ndim, length, sample_rate, fftlength, overlap, average, ndim, fast
):
    batch_size = 8
    num_channels = 5
    if overlap is not None and overlap >= fftlength:
        return

    if y_ndim == 1 and ndim == 1:
        return

    transform = SpectralDensity(
        sample_rate, fftlength, overlap, average=average, fast=fast
    )

    shape = [int(length * sample_rate)]
    if ndim > 1:
        shape.insert(0, num_channels)
    if ndim > 2:
        shape.insert(0, batch_size)
    x = np.random.randn(*shape)

    if ndim == 1 or (y_ndim == 1 and ndim == 2):
        y = np.random.randn(shape[-1])
    elif ndim == 3 and y_ndim == 1:
        y = np.random.randn(shape[0], shape[-1])
    else:
        y = np.random.randn(*shape)

    x = torch.Tensor(x)
    y = torch.Tensor(y)

    # make sure we catch if the fftlength is too long for the data
    if not fast:
        with pytest.raises(NotImplementedError):
            transform(x, y)
        return
    elif fftlength > length:
        with pytest.raises(ValueError) as exc_info:
            transform(x, y)
        assert str(exc_info.value).startswith("Number of samples")
        return

    # perform the transform and confirm the shape is correct
    torch_result = transform(x, y).numpy()
    num_freq_bins = int(fftlength * sample_rate) // 2 + 1
    shape[-1] = num_freq_bins
    assert torch_result.shape == tuple(shape)

    if ndim == 3:
        scipy_result = []
        if y_ndim == 1:
            x = x.transpose(1, 0)
            y = [y] * len(x)

        for i, j in zip(x, y):
            _, result = signal.csd(
                i,
                j,
                fs=sample_rate,
                nperseg=transform.nperseg,
                noverlap=transform.nperseg - transform.nstride,
                window=signal.windows.hann(transform.nperseg, False),
                average=average,
            )
            scipy_result.append(result)
        scipy_result = np.stack(scipy_result)

        if y_ndim == 1:
            x = x.transpose(1, 0)
            y = y[0]
            scipy_result = scipy_result.transpose(1, 0, 2)
    else:
        _, scipy_result = signal.csd(
            x,
            y,
            fs=sample_rate,
            nperseg=transform.nperseg,
            noverlap=transform.nperseg - transform.nstride,
            window=signal.windows.hann(transform.nperseg, False),
            average=average,
        )
    assert scipy_result.shape == torch_result.shape

    if fast:
        torch_result = torch_result[..., 2:]
        scipy_result = scipy_result[..., 2:]

    ratio = torch_result / scipy_result
    assert np.isclose(torch_result, scipy_result, rtol=TOL).all(), ratio

    _shape_checks(ndim, y_ndim, x, y, transform)



import numpy as np
import pytest
import torch
from gwpy.signal.qtransform import QPlane
from gwpy.signal.qtransform import QTile as gwpy_QTile
from gwpy.signal.qtransform import QTiling

from ml4gw.transforms import QScan, SingleQTransform
from ml4gw.transforms.qtransform import QTile


@pytest.fixture(params=[1, 2])
def duration(request):
    return request.param


@pytest.fixture(params=[1024, 2048])
def sample_rate(request):
    return request.param


@pytest.fixture(params=["mean", "median", None])
def norm(request):
    return request.param


@pytest.fixture(params=[[64, 64], [64, 128], [128, 128]])
def spectrogram_shape(request):
    return request.param


@pytest.fixture(params=[12, 50])
def q(request):
    return request.param


@pytest.fixture(params=[0.2, 0.35])
def mismatch(request):
    return request.param


@pytest.fixture(params=[128, 256])
def frequency(request):
    return request.param


@pytest.fixture(params=["bilinear", "bicubic", "spline"])
def interpolation_method(request):
    return request.param


def test_qtile(
    q,
    frequency,
    duration,
    sample_rate,
    mismatch,
    norm,
):
    X = torch.randn(int(duration * sample_rate))
    X = torch.fft.rfft(X, norm="forward")
    X[..., 1:] *= 2

    torch_qtile = QTile(q, frequency, duration, sample_rate, mismatch)
    gwpy_qtile = gwpy_QTile(q, frequency, duration, sample_rate, mismatch)

    assert torch_qtile.ntiles() == gwpy_qtile.ntiles
    assert np.allclose(
        torch_qtile.get_window().numpy(), gwpy_qtile.get_window()
    )
    assert (
        torch_qtile.get_data_indices().numpy() == gwpy_qtile.get_data_indices()
    ).all()
    assert np.allclose(
        torch_qtile(X, norm).numpy(),
        gwpy_qtile.transform(X, norm, 0),
        rtol=1e-3,
    )

    X = torch.randn(2, 2, 2, int(sample_rate * duration))
    with pytest.raises(ValueError):
        torch_qtile(X)


def test_singleqtransform(
    q,
    duration,
    sample_rate,
    mismatch,
    norm,
    spectrogram_shape,
    interpolation_method,
):
    X = torch.randn(int(duration * sample_rate))
    fseries = torch.fft.rfft(X, norm="forward")
    fseries[..., 1:] *= 2

    with pytest.raises(ValueError):
        qtransform = SingleQTransform(
            duration,
            sample_rate,
            spectrogram_shape,
            q,
            frange=[0, torch.inf],
            mismatch=mismatch,
            interpolation_method="nonsense",
        )

    with pytest.raises(ValueError):
        qtransform = SingleQTransform(
            duration,
            sample_rate,
            spectrogram_shape,
            q=1000,
            frange=[0, torch.inf],
            mismatch=mismatch,
            interpolation_method="nonsense",
        )

    qtransform = SingleQTransform(
        duration,
        sample_rate,
        spectrogram_shape,
        q,
        frange=[0, torch.inf],
        mismatch=mismatch,
        interpolation_method=interpolation_method,
    )

    with pytest.raises(RuntimeError):
        qtransform.get_max_energy()

    with pytest.raises(RuntimeError):
        qtransform.interpolate()

    qplane = QPlane(
        q,
        frange=[0, np.inf],
        duration=duration,
        sampling=sample_rate,
        mismatch=mismatch,
    )

    assert (
        qtransform.get_freqs().numpy() == list(qplane._iter_frequencies())
    ).all()

    qtransform.compute_qtiles(X, norm)
    torch_qtiles = qtransform.qtiles
    gwpy_qtiles = [qtile.transform(fseries, norm, 0) for qtile in qplane]

    for t, g in zip(torch_qtiles, gwpy_qtiles):
        assert np.allclose(t.numpy(), g, rtol=1e-3)

    transformed = qtransform(X, norm)
    assert list(transformed.shape[-2:]) == spectrogram_shape


def test_get_qs(
    duration,
    sample_rate,
    mismatch,
    spectrogram_shape,
):
    frange = [0, torch.inf]
    qrange = [1, 1000]

    qscan = QScan(
        duration,
        sample_rate,
        spectrogram_shape,
        qrange,
        frange,
        mismatch=mismatch,
    )
    qtiling = QTiling(
        duration, sample_rate, qrange, frange=[0, np.inf], mismatch=mismatch
    )

    assert np.allclose(qscan.get_qs(), qtiling.qs)

    # Just check that the QScan runs
    data = torch.randn(int(sample_rate * duration))
    _ = qscan(data)



import pytest
import torch
from torchaudio.transforms import Spectrogram

from ml4gw.transforms import MultiResolutionSpectrogram


@pytest.fixture(params=[2, 4, 5])
def kernel_length(request):
    return request.param


@pytest.fixture(params=[128, 256])
def sample_rate(request):
    return request.param


@pytest.fixture(params=[1, 10])
def batch_size(request):
    return request.param


@pytest.fixture(params=[1, 3])
def num_channels(request):
    return request.param


@pytest.fixture(params=[[64], [64, 128], [64, 128, 256]])
def n_ffts(request):
    return request.param


@pytest.fixture(params=[[50]])
def win_lengths(request):
    return request.param


@pytest.fixture(params=[[2]])
def powers(request):
    return request.param


def test_num_spectrogram_warning():
    with pytest.warns(RuntimeWarning):
        MultiResolutionSpectrogram(8, 2048, n_fft=[32, 64, 128, 256])


def test_normalization_error():
    with pytest.raises(ValueError):
        MultiResolutionSpectrogram(8, 2048, normalized=[False])


def test_arg_length_error():
    with pytest.raises(ValueError):
        MultiResolutionSpectrogram(8, 2048, n_fft=[32, 64], power=[1, 1, 1])


def test_kernel_size_error():
    with pytest.raises(ValueError):
        spectrogram = MultiResolutionSpectrogram(8, 2048)
        spectrogram(torch.ones([4, 3, 8 * 2048 + 1]))


def test_multi_resolution_spectrogram(
    kernel_length,
    sample_rate,
    batch_size,
    num_channels,
    n_ffts,
    win_lengths,
    powers,
):
    # Creating a MRS without any spectrogram arguments should
    # just create a single default torchaudio histogram with
    # `normalized = True`
    spectrogram = MultiResolutionSpectrogram(kernel_length, sample_rate)

    X = torch.ones([batch_size, num_channels, kernel_length * sample_rate])
    y = spectrogram(X)
    expected_y = Spectrogram(normalized=True)(X)

    assert (y == expected_y).all()

    # Check that all the indexing we're doing is working by
    # performing a more explicit version of the algorithm
    spectrogram = MultiResolutionSpectrogram(
        kernel_length,
        sample_rate,
        n_fft=n_ffts,
        win_length=win_lengths,
        power=powers,
    )
    y = spectrogram(X)
    kwargs = spectrogram.kwargs
    ta_spectrograms = [Spectrogram(**k)(X[0, 0]) for k in kwargs]
    t_dim = max([spec.shape[-1] for spec in ta_spectrograms])
    f_dim = max([spec.shape[-2] for spec in ta_spectrograms])
    expected_y = torch.zeros([f_dim, t_dim])

    for i in range(t_dim):
        for j in range(f_dim):
            max_value = 0
            for spec in ta_spectrograms:
                t_idx = int(i / t_dim * spec.shape[-1])
                f_idx = int(j / f_dim * spec.shape[-2])
                max_value = max(max_value, spec[f_idx, t_idx])
            expected_y[j, i] += max_value

    assert torch.allclose(y[0, 0], expected_y, rtol=1e-6)



import random

import pytest
import torch

from ml4gw.dataloading import ChunkedTimeSeriesDataset


class TestChunkedTimeseriesDataset:
    @pytest.fixture
    def chunk_length(self):
        return 2

    @pytest.fixture
    def chunks_per_epoch(self):
        return 6

    @pytest.fixture
    def chunk_it(self, chunk_length, chunks_per_epoch):
        def it():
            for _ in range(chunks_per_epoch):
                chunk = []
                for _ in range(6):
                    start = random.randint(0, 128 * 8)
                    stop = start + 128 * chunk_length
                    row = torch.arange(start, stop)
                    row = torch.stack([row, -row])
                    chunk.append(row)
                chunk = torch.stack(chunk)
                yield chunk

        return it

    @pytest.fixture
    def kernel_length(self):
        return 1.5

    @pytest.fixture
    def batch_size(self):
        return 8

    @pytest.fixture
    def batches_per_chunk(self):
        return 7

    @pytest.fixture(params=[True, False])
    def coincident(self, request):
        return request.param

    @pytest.fixture
    def dataset(
        self,
        chunk_it,
        kernel_length,
        batch_size,
        batches_per_chunk,
        coincident,
    ):
        return ChunkedTimeSeriesDataset(
            chunk_it(),
            int(kernel_length * 128),
            batch_size=batch_size,
            batches_per_chunk=batches_per_chunk,
            coincident=coincident,
            device="cpu",
        )

    def test_iter(self, dataset, coincident):
        length = 0
        for x in dataset:
            assert x.shape == (8, 2, 192)
            if coincident:
                torch.testing.assert_close(x[:, 0], -x[:, 1], rtol=0, atol=0)
            else:
                assert (x[:, 0] != -x[:, 1]).any().item()

            diffs = torch.diff(x[:, 0], axis=-1)
            expected = torch.ones_like(diffs)
            torch.testing.assert_close(diffs, expected, rtol=0, atol=0)
            length += 1

        assert length == 42



import lal
import lalsimulation
import numpy as np
import pytest
import torch
from astropy import units as u

import ml4gw.waveforms as waveforms
from ml4gw.waveforms.conversion import (
    bilby_spins_to_lalsim,
    chirp_mass_and_mass_ratio_to_components,
)


@pytest.fixture()
def num_samples():
    return 100


@pytest.fixture(params=[20, 40])
def f_ref(request):
    return request.param


def test_taylor_f2(
    chirp_mass,
    mass_ratio,
    chi1,
    chi2,
    phase,
    distance,
    f_ref,
    theta_jn,
    sample_rate,
):
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )

    # compare each waveform with lalsimulation
    for i in range(len(chirp_mass)):
        # construct lalinference params
        params = {
            "m1": mass_1[i].item() * lal.MSUN_SI,
            "m2": mass_2[i].item() * lal.MSUN_SI,
            "S1x": 0,
            "S1y": 0,
            "S1z": chi1[i].item(),
            "S2x": 0,
            "S2y": 0,
            "S2z": chi2[i].item(),
            "distance": (distance[i].item() * u.Mpc).to("m").value,
            "inclination": theta_jn[i].item(),
            "phiRef": phase[i].item(),
            "longAscNodes": 0.0,
            "eccentricity": 0.0,
            "meanPerAno": 0.0,
            "deltaF": 1.0 / sample_rate,
            "f_min": 10,
            "f_ref": f_ref,
            "f_max": 300,
            "approximant": lalsimulation.TaylorF2,
            "LALpars": lal.CreateDict(),
        }
        hp_lal, hc_lal = lalsimulation.SimInspiralChooseFDWaveform(**params)

        # reconstruct frequencies generated by
        # lal and filter based on fmin and fmax
        lal_freqs = np.array(
            [
                hp_lal.f0 + ii * hp_lal.deltaF
                for ii in range(len(hp_lal.data.data))
            ]
        )

        lal_mask = (lal_freqs > params["f_min"]) & (
            lal_freqs < params["f_max"]
        )

        lal_freqs = lal_freqs[lal_mask]
        torch_freqs = torch.tensor(lal_freqs, dtype=torch.float64)

        # generate waveforms using ml4gw
        hc_ml4gw, hp_ml4gw = waveforms.TaylorF2()(
            torch_freqs,
            chirp_mass[i][None],
            mass_ratio[i][None],
            chi1[i][None],
            chi2[i][None],
            distance[i][None],
            phase[i][None],
            theta_jn[i][None],
            f_ref,
        )

        hc_ml4gw = hc_ml4gw[0]
        hp_ml4gw = hp_ml4gw[0]

        hp_lal_data = hp_lal.data.data[lal_mask]
        hc_lal_data = hc_lal.data.data[lal_mask]

        # ensure no nans
        assert not torch.any(torch.isnan(hc_ml4gw))
        assert not torch.any(torch.isnan(hp_ml4gw))

        assert np.allclose(
            1e21 * hp_lal_data.real, 1e21 * hp_ml4gw.real.numpy(), atol=1e-3
        )
        assert np.allclose(
            1e21 * hp_lal_data.imag, 1e21 * hp_ml4gw.imag.numpy(), atol=1e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.real, 1e21 * hc_ml4gw.real.numpy(), atol=1e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.imag, 1e21 * hc_ml4gw.imag.numpy(), atol=1e-3
        )

        # taylor f2 is symmetric w.r.t m1 --> m2 flip.
        # so test that the waveforms are the same when m1 and m2
        # (and corresponding chi1, chi2 are flipped)
        # are flipped this can be done by flipping mass ratio
        hc_ml4gw, hp_ml4gw = waveforms.TaylorF2()(
            torch_freqs,
            chirp_mass[i][None],
            1 / mass_ratio[i][None],
            chi2[i][None],
            chi1[i][None],
            distance[i][None],
            phase[i][None],
            theta_jn[i][None],
            f_ref,
        )

        hc_ml4gw = hc_ml4gw[0]
        hp_ml4gw = hp_ml4gw[0]

        assert np.allclose(
            1e21 * hp_lal_data.real, 1e21 * hp_ml4gw.real.numpy(), atol=1e-3
        )
        assert np.allclose(
            1e21 * hp_lal_data.imag, 1e21 * hp_ml4gw.imag.numpy(), atol=1e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.real, 1e21 * hc_ml4gw.real.numpy(), atol=1e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.imag, 1e21 * hc_ml4gw.imag.numpy(), atol=1e-3
        )


def test_phenom_d(
    chirp_mass,
    mass_ratio,
    chi1,
    chi2,
    distance,
    phase,
    theta_jn,
    sample_rate,
    f_ref,
):
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )

    # compare each waveform with lalsimulation
    for i in range(len(chirp_mass)):
        # construct lalinference params
        params = {
            "m1": mass_1[i].item() * lal.MSUN_SI,
            "m2": mass_2[i].item() * lal.MSUN_SI,
            "S1x": 0,
            "S1y": 0,
            "S1z": chi1[i].item(),
            "S2x": 0,
            "S2y": 0,
            "S2z": chi2[i].item(),
            "distance": (distance[i].item() * u.Mpc).to("m").value,
            "inclination": theta_jn[i].item(),
            "phiRef": phase[i].item(),
            "longAscNodes": 0.0,
            "eccentricity": 0.0,
            "meanPerAno": 0.0,
            "deltaF": 1.0 / sample_rate,
            "f_min": 10,
            "f_ref": f_ref,
            "f_max": 300,
            "approximant": lalsimulation.IMRPhenomD,
            "LALpars": lal.CreateDict(),
        }
        hp_lal, hc_lal = lalsimulation.SimInspiralChooseFDWaveform(**params)

        # reconstruct frequencies generated by
        # lal and filter based on fmin and fmax
        lal_freqs = np.array(
            [
                hp_lal.f0 + ii * hp_lal.deltaF
                for ii in range(len(hp_lal.data.data))
            ]
        )

        lal_mask = (lal_freqs > params["f_min"]) & (
            lal_freqs < params["f_max"]
        )

        lal_freqs = lal_freqs[lal_mask]
        torch_freqs = torch.tensor(lal_freqs, dtype=torch.float32)

        # generate waveforms using ml4gw
        hc_ml4gw, hp_ml4gw = waveforms.IMRPhenomD()(
            torch_freqs,
            chirp_mass[i][None],
            mass_ratio[i][None],
            chi1[i][None],
            chi2[i][None],
            distance[i][None],
            phase[i][None],
            theta_jn[i][None],
            f_ref,
        )

        hc_ml4gw = hc_ml4gw[0]
        hp_ml4gw = hp_ml4gw[0]

        hp_lal_data = hp_lal.data.data[lal_mask]
        hc_lal_data = hc_lal.data.data[lal_mask]

        assert not torch.any(torch.isnan(hc_ml4gw))
        assert not torch.any(torch.isnan(hp_ml4gw))

        assert np.allclose(
            1e21 * hp_lal_data.real, 1e21 * hp_ml4gw.real.numpy(), atol=2e-3
        )
        assert np.allclose(
            1e21 * hp_lal_data.imag, 1e21 * hp_ml4gw.imag.numpy(), atol=2e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.real, 1e21 * hc_ml4gw.real.numpy(), atol=2e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.imag, 1e21 * hc_ml4gw.imag.numpy(), atol=2e-3
        )


def test_phenom_p(
    chirp_mass,
    mass_ratio,
    distance_far,
    distance_close,
    phase,
    sample_rate,
    f_ref,
    theta_jn,
    phi_jl,
    tilt_1,
    tilt_2,
    phi_12,
    a_1,
    a_2,
):
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )
    (
        inclination,
        chi1x,
        chi1y,
        chi1z,
        chi2x,
        chi2y,
        chi2z,
    ) = bilby_spins_to_lalsim(
        theta_jn,
        phi_jl,
        tilt_1,
        tilt_2,
        phi_12,
        a_1,
        a_2,
        mass_1,
        mass_2,
        f_ref,
        phase,
    )

    tc = 0.0

    # compare each waveform with lalsimulation
    for i in range(len(chirp_mass)):
        # test far (> 400 Mpc) waveforms (O(1e-3) agreement)

        # construct lalinference params
        params = {
            "m1": mass_1[i].item() * lal.MSUN_SI,
            "m2": mass_2[i].item() * lal.MSUN_SI,
            "S1x": chi1x[i].item(),
            "S1y": chi1y[i].item(),
            "S1z": chi1z[i].item(),
            "S2x": chi2x[i].item(),
            "S2y": chi2y[i].item(),
            "S2z": chi2z[i].item(),
            "distance": (distance_far[i].item() * u.Mpc).to("m").value,
            "inclination": inclination[i].item(),
            "phiRef": phase[i].item(),
            "longAscNodes": 0.0,
            "eccentricity": 0.0,
            "meanPerAno": 0.0,
            "deltaF": 1.0 / sample_rate,
            "f_min": 10.0,
            "f_ref": f_ref,
            "f_max": 300,
            "approximant": lalsimulation.IMRPhenomPv2,
            "LALpars": lal.CreateDict(),
        }
        hp_lal, hc_lal = lalsimulation.SimInspiralChooseFDWaveform(**params)

        # reconstruct frequencies generated by
        # lal and filter based on fmin and fmax
        lal_freqs = np.array(
            [
                hp_lal.f0 + ii * hp_lal.deltaF
                for ii in range(len(hp_lal.data.data))
            ]
        )

        lal_mask = (lal_freqs > params["f_min"]) & (
            lal_freqs < params["f_max"]
        )

        lal_freqs = lal_freqs[lal_mask]
        torch_freqs = torch.tensor(lal_freqs, dtype=torch.float32)

        hc_ml4gw, hp_ml4gw = waveforms.IMRPhenomPv2()(
            torch_freqs,
            chirp_mass[i][None],
            mass_ratio[i][None],
            chi1x[i][None],
            chi1y[i][None],
            chi1z[i][None],
            chi2x[i][None],
            chi2y[i][None],
            chi2z[i][None],
            distance_far[i][None],
            phase[i][None],
            inclination[i][None],
            f_ref,
            torch.tensor([tc]),
        )

        hp_ml4gw = hp_ml4gw[0]
        hc_ml4gw = hc_ml4gw[0]

        hp_lal_data = hp_lal.data.data[lal_mask]
        hc_lal_data = hc_lal.data.data[lal_mask]

        # Only 4 of 50,000 samples failed this tolerance
        assert np.allclose(
            1e21 * hp_lal_data.real, 1e21 * hp_ml4gw.real.numpy(), atol=2e-3
        )
        assert np.allclose(
            1e21 * hp_lal_data.imag, 1e21 * hp_ml4gw.imag.numpy(), atol=2e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.real, 1e21 * hc_ml4gw.real.numpy(), atol=2e-3
        )
        assert np.allclose(
            1e21 * hc_lal_data.imag, 1e21 * hc_ml4gw.imag.numpy(), atol=2e-3
        )

        # test close (< 400 Mpc) waveforms  (O(1e-2) agreement)
        params["distance"] = (distance_close[i].item() * u.Mpc).to("m").value
        hp_lal, hc_lal = lalsimulation.SimInspiralChooseFDWaveform(**params)

        # reconstruct frequencies generated by
        # lal and filter based on fmin and fmax
        lal_freqs = np.array(
            [
                hp_lal.f0 + ii * hp_lal.deltaF
                for ii in range(len(hp_lal.data.data))
            ]
        )

        lal_mask = (lal_freqs > params["f_min"]) & (
            lal_freqs < params["f_max"]
        )

        lal_freqs = lal_freqs[lal_mask]
        torch_freqs = torch.tensor(lal_freqs, dtype=torch.float32)

        hc_ml4gw, hp_ml4gw = waveforms.IMRPhenomPv2()(
            torch_freqs,
            chirp_mass[i][None],
            mass_ratio[i][None],
            chi1x[i][None],
            chi1y[i][None],
            chi1z[i][None],
            chi2x[i][None],
            chi2y[i][None],
            chi2z[i][None],
            distance_close[i][None],
            phase[i][None],
            inclination[i][None],
            f_ref,
            torch.tensor([tc]),
        )

        hp_ml4gw = hp_ml4gw[0]
        hc_ml4gw = hc_ml4gw[0]

        hp_lal_data = hp_lal.data.data[lal_mask]
        hc_lal_data = hc_lal.data.data[lal_mask]

        assert np.allclose(
            1e21 * hp_lal_data.real, 1e21 * hp_ml4gw.real.numpy(), atol=2e-2
        )
        assert np.allclose(
            1e21 * hp_lal_data.imag, 1e21 * hp_ml4gw.imag.numpy(), atol=2e-2
        )
        assert np.allclose(
            1e21 * hc_lal_data.real, 1e21 * hc_ml4gw.real.numpy(), atol=2e-2
        )
        assert np.allclose(
            1e21 * hc_lal_data.imag, 1e21 * hc_ml4gw.imag.numpy(), atol=2e-2
        )

    # test batched outputs works as expected
    hc_ml4gw, hp_ml4gw = waveforms.IMRPhenomPv2()(
        torch_freqs,
        chirp_mass[-1][None].repeat(10),
        mass_ratio[-1][None].repeat(10),
        chi1x[-1][None].repeat(10),
        chi1y[-1][None].repeat(10),
        chi1z[-1][None].repeat(10),
        chi2x[-1][None].repeat(10),
        chi2y[-1][None].repeat(10),
        chi2z[-1][None].repeat(10),
        distance_close[-1][None].repeat(10),
        phase[-1][None].repeat(10),
        inclination[-1][None].repeat(10),
        f_ref,
        torch.tensor([tc]).repeat(10),
    )

    # check batch against lal
    assert np.allclose(
        1e21 * hp_lal_data.real, 1e21 * hp_ml4gw[0].real.numpy(), atol=2e-2
    )
    assert np.allclose(
        1e21 * hp_lal_data.imag, 1e21 * hp_ml4gw[0].imag.numpy(), atol=2e-2
    )
    assert np.allclose(
        1e21 * hc_lal_data.real, 1e21 * hc_ml4gw[0].real.numpy(), atol=2e-2
    )
    assert np.allclose(
        1e21 * hc_lal_data.imag, 1e21 * hc_ml4gw[0].imag.numpy(), atol=2e-2
    )

    # check batch against each other
    for i in range(9):
        assert np.allclose(
            1e21 * hp_ml4gw[0].real.numpy(),
            1e21 * hp_ml4gw[i + 1].real.numpy(),
            atol=1e-2,
        )
        assert np.allclose(
            1e21 * hp_ml4gw[0].imag.numpy(),
            1e21 * hp_ml4gw[i + 1].imag.numpy(),
            atol=1e-2,
        )
        assert np.allclose(
            1e21 * hc_ml4gw[0].real.numpy(),
            1e21 * hc_ml4gw[i + 1].real.numpy(),
            atol=1e-2,
        )
        assert np.allclose(
            1e21 * hc_ml4gw[0].imag.numpy(),
            1e21 * hc_ml4gw[i + 1].imag.numpy(),
            atol=1e-2,
        )



from unittest.mock import patch

import numpy as np
import pytest
import torch

from ml4gw.utils import slicing


@pytest.fixture(params=[1, 8, 9])
def kernel_size(request):
    return request.param


@pytest.fixture(params=[1, 5])
def num_channels(request):
    return request.param


def test_unfold_windows():
    # 1D
    x = torch.tensor([1, 2, 3, 4, 5, 6], dtype=float)
    result = slicing.unfold_windows(x, window_size=3, stride=2)
    assert result.tolist() == [[1, 2, 3], [3, 4, 5]]

    result, rem = slicing.unfold_windows(
        x, window_size=3, stride=2, drop_last=False
    )
    assert rem.tolist() == [[6]]

    # 2D
    x = torch.tensor([[1, 2, 3, 4, 5, 6], [2, 3, 4, 5, 6, 7]], dtype=float)
    result = slicing.unfold_windows(x, window_size=3, stride=2)
    assert result.tolist() == [[[1, 2, 3], [2, 3, 4]], [[3, 4, 5], [4, 5, 6]]]

    result, rem = slicing.unfold_windows(
        x, window_size=3, stride=2, drop_last=False
    )
    assert rem.tolist() == [[[6], [7]]]

    # 3D
    x = torch.tensor(
        [
            [[1, 2, 3, 4, 5, 6], [2, 3, 4, 5, 6, 7]],
            [[3, 4, 5, 6, 7, 8], [4, 5, 6, 7, 8, 9]],
        ],
        dtype=float,
    )
    result = slicing.unfold_windows(x, window_size=3, stride=2)
    assert result.tolist() == [
        [[[1, 2, 3], [2, 3, 4]], [[3, 4, 5], [4, 5, 6]]],
        [[[3, 4, 5], [4, 5, 6]], [[5, 6, 7], [6, 7, 8]]],
    ]

    result, rem = slicing.unfold_windows(
        x, window_size=3, stride=2, drop_last=False
    )
    assert rem.tolist() == [[[[6], [7]], [[8], [9]]]]


def test_slice_kernels(kernel_size, num_channels):
    x = torch.arange(100)
    idx = 7 * torch.arange(2, 10)

    # test 1D slice
    result = slicing.slice_kernels(x, idx, kernel_size)
    assert result.shape == (8, kernel_size)
    for i, y in enumerate(result.cpu().numpy()):
        start = (i + 2) * 7
        stop = start + kernel_size
        assert (y == np.arange(start, stop)).all()

    # ensure 1D slice with >1D idx raises error
    with pytest.raises(ValueError) as exc:
        slicing.slice_kernels(x, idx[None], kernel_size)
    assert str(exc.value).startswith("idx tensor has 2 dimensions")

    # test 2D slice with 1D indices
    X = torch.stack([x + i * 100 for i in range(num_channels)])
    result = slicing.slice_kernels(X, idx, kernel_size)
    assert result.shape == (8, num_channels, kernel_size)

    for i, Y in enumerate(result.cpu().numpy()):
        for j, y in enumerate(Y):
            start = j * 100 + (i + 2) * 7
            stop = start + kernel_size
            assert (y == np.arange(start, stop)).all()

    # test 2D slice with 2D indices
    idx = torch.stack([idx + i * 3 for i in range(num_channels)]).t()
    result = slicing.slice_kernels(X, idx, kernel_size)
    assert result.shape == (8, num_channels, kernel_size)

    for i, Y in enumerate(result.cpu().numpy()):
        for j, y in enumerate(Y):
            start = j * 103 + (i + 2) * 7
            stop = start + kernel_size
            assert (y == np.arange(start, stop)).all()

    # ensure that 2D slice with > 2D indices raises error
    idx = idx[:, :, None]
    with pytest.raises(ValueError) as exc:
        slicing.slice_kernels(X, idx, kernel_size)
    assert str(exc.value).startswith("Can't slice 2D array")

    # test 3D slice with 1D indices
    idx = idx[:, 0, 0]
    X = torch.stack([X + i * 1000 for i in range(len(idx))])
    result = slicing.slice_kernels(X, idx, kernel_size)
    for i, Y in enumerate(result.cpu().numpy()):
        for j, y in enumerate(Y):
            start = i * 1007 + j * 100 + 14
            stop = start + kernel_size
            assert (y == np.arange(start, stop)).all(), (i, j)

    # ensure 3D slice with wrong number of indices raises error
    with pytest.raises(ValueError) as exc:
        slicing.slice_kernels(X, idx[:-2], kernel_size)
    assert str(exc.value).startswith("Can't slice kernels from batch")

    # ensure 3D slice with >1D indices raises error
    idx = idx[:, None]
    with pytest.raises(ValueError) as exc:
        slicing.slice_kernels(X, idx, kernel_size)
    assert str(exc.value).startswith("Can't slice 3D array")

    # ensure >3D slice raises error
    idx = idx[:, 0]
    X = X[None]
    with pytest.raises(ValueError) as exc:
        slicing.slice_kernels(X, idx, kernel_size)
    assert str(exc.value).startswith("Can't slice kernels from tensor")


def test_sample_kernels_1D(kernel_size, num_channels):
    N = 8
    x = torch.arange(100)

    # make sure that we enforce that we have enough data to sample
    with pytest.raises(ValueError) as exc:
        slicing.sample_kernels(x[: kernel_size - 1], kernel_size, N)
    assert str(exc.value).startswith("Can't sample kernels of size")

    # make sure that we enforce that N is not None
    with pytest.raises(ValueError) as exc:
        slicing.sample_kernels(x, kernel_size)
    assert str(exc.value).startswith("Must specify number of kernels")

    # first test 1D behavior
    return_value = 7 * torch.arange(N)
    with patch("torch.randint", return_value=return_value) as mock:
        result = slicing.sample_kernels(x, kernel_size, N)
    mock.assert_called_once_with(100 - kernel_size, size=(N,))

    assert result.shape == (N, kernel_size)
    for i, y in enumerate(result.cpu().numpy()):
        start = i * 7
        stop = start + kernel_size
        assert (y == np.arange(start, stop)).all()

    # now 1D behavior without patching
    result = slicing.sample_kernels(x, kernel_size, N)
    assert result.shape == (N, kernel_size)
    for y in result.cpu().numpy():
        assert y[0] <= (100 - kernel_size)
        assert (y == np.arange(y[0], y[0] + kernel_size)).all()


@pytest.fixture(params=[None, 0, 4, -4])
def max_center_offset(request):
    return request.param


@pytest.fixture(params=[True, False])
def coincident(request):
    return request.param


def test_sample_kernels_2D(
    kernel_size, num_channels, max_center_offset, coincident
):
    N = 8
    x = torch.arange(100)
    X = torch.stack([x + i * 100 for i in range(num_channels)])

    # make sure that we enforce that we have enough data to sample
    with pytest.raises(ValueError) as exc:
        slicing.sample_kernels(X[:, : kernel_size - 1], kernel_size, N)
    assert str(exc.value).startswith("Can't sample kernels of size")

    # make sure that we enforce that N is not None
    with pytest.raises(ValueError) as exc:
        slicing.sample_kernels(X, kernel_size)
    assert str(exc.value).startswith("Must specify number of kernels")

    # make sure we enforce that min_val is not negative
    if max_center_offset is not None:
        with pytest.raises(ValueError) as exc:
            bad_kernel_size = int(X.shape[-1] // 2) - max_center_offset + 1
            slicing.sample_kernels(X, bad_kernel_size, N, max_center_offset)
        assert str(exc.value).startswith("Kernel size")

    # for non-coincident sampling, we'll request a 2D
    # matrix of sample indices from `torch.randint`
    # rather than 1, and we'll make sure to create a
    # return value that has different values for each
    # channel to ensure the correct response
    shape = (N,)
    return_value = 7 * torch.arange(N)
    if not coincident:
        return_value = torch.stack(
            [return_value + i for i in range(num_channels)]
        ).t()
        shape = (N, num_channels)

    # if we requested some padding between the
    # edge of the kernel and the center of the
    # timeseries, but our kernel isn't big enough
    # to accommodate it, this should raise an error
    if (
        max_center_offset is not None
        and max_center_offset < 0
        and kernel_size <= 2 * abs(max_center_offset)
    ):
        with pytest.raises(ValueError):
            slicing.sample_kernels(
                X, kernel_size, N, max_center_offset, coincident
            )
        return

    with patch("torch.randint", return_value=return_value) as mock:
        result = slicing.sample_kernels(
            X, kernel_size, N, max_center_offset, coincident
        )
    assert result.shape == (N, num_channels, kernel_size)
    result = result.cpu().numpy()

    if max_center_offset is None:
        min_val, max_val = 0, 100 - kernel_size
    elif max_center_offset == 0:
        min_val, max_val = 50 - kernel_size, 50
    elif max_center_offset == -4:
        min_val, max_val = 54 - kernel_size, 46
    else:
        min_val, max_val = 46 - kernel_size, 50
    mock.assert_called_with(min_val, max_val, size=shape)

    for i, sample in enumerate(result):
        for j, channel in enumerate(sample):
            if coincident:
                start = i * 7 + j * 100
                stop = start + kernel_size
            else:
                start = i * 7 + j * 101
                stop = start + kernel_size

            assert (channel == np.arange(start, stop)).all()


@pytest.fixture(params=[None, 1])
def N(request):
    return request.param


def test_sample_kernels_3D(kernel_size, num_channels, max_center_offset, N):
    batch_size = 8
    if N is not None:
        N = batch_size

    x = torch.arange(100)
    X = torch.stack([x + i * 100 for i in range(num_channels)])
    X = torch.stack([X + i * 1000 for i in range(batch_size)])

    # make sure that we enforce that we have enough data to sample
    with pytest.raises(ValueError) as exc:
        slicing.sample_kernels(X[:, :, : kernel_size - 1], kernel_size, N)
    assert str(exc.value).startswith("Can't sample kernels of size")

    # make sure we raise an error if we specify N
    # but it doesn't match the batch size of X
    with pytest.raises(ValueError) as exc:
        slicing.sample_kernels(X, kernel_size, batch_size - 1)
    assert str(exc.value).startswith(f"Can't sample {batch_size - 1} kernels")

    # make sure we enforce that min_val is not negative
    if max_center_offset is not None:
        with pytest.raises(ValueError) as exc:
            bad_kernel_size = int(X.shape[-1] // 2) - max_center_offset + 1
            slicing.sample_kernels(X, bad_kernel_size, N, max_center_offset)
        assert str(exc.value).startswith("Kernel size")

    # if we requested some padding between the
    # edge of the kernel and the center of the
    # timeseries, but our kernel isn't big enough
    # to accommodate it, this should raise an error
    if (
        max_center_offset is not None
        and max_center_offset < 0
        and kernel_size <= 2 * abs(max_center_offset)
    ):
        with pytest.raises(ValueError):
            slicing.sample_kernels(
                X, kernel_size, N, max_center_offset, coincident
            )
        return

    return_value = 7 * torch.arange(batch_size)
    with patch("torch.randint", return_value=return_value) as mock:
        result = slicing.sample_kernels(
            X, kernel_size, N, max_center_offset, coincident
        )
    assert result.shape == (batch_size, num_channels, kernel_size)
    result = result.cpu().numpy()

    if max_center_offset is None:
        min_val, max_val = 0, 100 - kernel_size
    elif max_center_offset == 0:
        min_val, max_val = 50 - kernel_size, 50
    elif max_center_offset == -4:
        min_val, max_val = 54 - kernel_size, 46
    else:
        min_val, max_val = 46 - kernel_size, 50
    mock.assert_called_with(min_val, max_val, size=(batch_size,))

    for i, sample in enumerate(result):
        for j, channel in enumerate(sample):
            if coincident:
                start = i * 1007 + j * 100
                stop = start + kernel_size
            else:
                start = i * 1007 + j * 101
                stop = start + kernel_size

            assert (channel == np.arange(start, stop)).all()



import numpy as np
import torch
from lalsimulation import SimInspiralTransformPrecessingNewInitialConditions
from torch.distributions import Uniform

from ml4gw.constants import MSUN
from ml4gw.waveforms.conversion import bilby_spins_to_lalsim


def test_bilby_to_lalsim_spins():
    theta_jn = Uniform(0, torch.pi).sample((100,))
    phi_jl = Uniform(0, 2 * torch.pi).sample((100,))
    tilt_1 = Uniform(0, torch.pi).sample((100,))
    tilt_2 = Uniform(0, torch.pi).sample((100,))
    phi_12 = Uniform(0, 2 * torch.pi).sample((100,))
    a_1 = Uniform(0, 0.99).sample((100,))
    a_2 = Uniform(0, 0.99).sample((100,))
    mass_1 = Uniform(3, 100).sample((100,))
    mass_2 = Uniform(3, 100).sample((100,))
    f_ref = 40.0
    phi_ref = Uniform(0, torch.pi).sample((100,))
    incl, s1x, s1y, s1z, s2x, s2y, s2z = bilby_spins_to_lalsim(
        theta_jn,
        phi_jl,
        tilt_1,
        tilt_2,
        phi_12,
        a_1,
        a_2,
        mass_1,
        mass_2,
        f_ref,
        phi_ref,
    )
    for i in range(2):
        (
            lal_incl,
            lal_s1x,
            lal_s1y,
            lal_s1z,
            lal_s2x,
            lal_s2y,
            lal_s2z,
        ) = SimInspiralTransformPrecessingNewInitialConditions(
            theta_jn[i].item(),
            phi_jl[i].item(),
            tilt_1[i].item(),
            tilt_2[i].item(),
            phi_12[i].item(),
            a_1[i].item(),
            a_2[i].item(),
            mass_1[i].item() * MSUN,
            mass_2[i].item() * MSUN,
            f_ref,
            phi_ref[i].item(),
        )

        # check if the values are close up to 4 decimal places
        assert np.isclose(incl[i].item(), lal_incl, atol=1e-4)
        assert np.isclose(s1x[i].item(), lal_s1x, atol=1e-4)
        assert np.isclose(s1y[i].item(), lal_s1y, atol=1e-4)
        assert np.isclose(s1z[i].item(), lal_s1z, atol=1e-4)
        assert np.isclose(s2x[i].item(), lal_s2x, atol=1e-4)
        assert np.isclose(s2y[i].item(), lal_s2y, atol=1e-4)
        assert np.isclose(s2z[i].item(), lal_s2z, atol=1e-4)



import pytest
import torch

from ml4gw.nn.norm import GroupNorm1D


@pytest.fixture(params=[1, 2, 3, 4])
def factor(request):
    return request.param


@pytest.fixture(params=[None, 1, 2, 3, 4])
def num_groups(request):
    return request.param


def update_weights(norm):
    optim = torch.optim.SGD(norm.parameters(), lr=1e-1)
    for _ in range(10):
        optim.zero_grad()
        x, y = [torch.randn(8, norm.num_channels, 128) for _ in range(2)]
        y = 0.2 + 0.5 * y
        y_hat = norm(x)
        loss = ((y_hat - y) ** 2).mean()
        loss.backward()
        optim.step()
    return norm


def copy_weights(target, source):
    target.weight.requires_grad = False
    target.bias.requires_grad = False
    target.weight.copy_(source.weight.data[:, 0])
    target.bias.copy_(source.bias.data[:, 0])
    return target


def test_group_norm(num_groups, factor):
    with pytest.raises(ValueError):
        GroupNorm1D(1, 2)

    with pytest.raises(ValueError):
        GroupNorm1D(5, 3)

    if num_groups is None:
        num_channels = factor
    else:
        num_channels = num_groups * factor

    norm = GroupNorm1D(num_channels, num_groups)

    with pytest.raises(ValueError):
        x = torch.randn(num_channels, 1024)
        x = norm(x)

    with pytest.raises(ValueError):
        x = torch.randn(1, 128, num_channels, 1024)
        x = norm(x)

    # update the norm layers weights so that
    # we have something interesting to compare
    norm = update_weights(norm)

    # copy learned parameters into normal groupnorm
    # and verify that outputs are similar
    ref = torch.nn.GroupNorm(norm.num_groups, norm.num_channels)
    ref = copy_weights(ref, norm)

    x = torch.randn(128, num_channels, 1024)
    x_ref = ref(x)
    x = norm(x)

    close = torch.isclose(x, x_ref, rtol=1e-6)
    num_wrong = (~close).sum()
    assert (num_wrong / x.numel()) < 0.01



import h5py  # noqa
import pytest
import torch

from ml4gw.nn.resnet.resnet_2d import (
    BasicBlock,
    Bottleneck,
    BottleneckResNet2D,
    ResNet2D,
    conv1,
)


@pytest.fixture(params=[3, 7, 8])
def kernel_size(request):
    return request.param


@pytest.fixture(params=[10, 50])
def spectrogram_size(request):
    return request.param


@pytest.fixture(params=[1, 2])
def stride(request):
    return request.param


@pytest.fixture(params=[2, 4])
def inplanes(request):
    return request.param


@pytest.fixture(params=[1, 2])
def classes(request):
    return request.param


@pytest.fixture(params=[BasicBlock, Bottleneck])
def block(request):
    return request.param


def test_blocks(block, kernel_size, stride, spectrogram_size, inplanes):
    # TODO: test dilation for bottleneck
    planes = 4

    if stride > 1 or inplanes != planes * block.expansion:
        downsample = conv1(inplanes, planes * block.expansion, stride)
    else:
        downsample = None

    if kernel_size % 2 == 0:
        with pytest.raises(ValueError):
            block = block(
                inplanes, planes, kernel_size, stride, downsample=downsample
            )
        return

    block = block(inplanes, planes, kernel_size, stride, downsample=downsample)
    x = torch.randn(8, inplanes, spectrogram_size, spectrogram_size)
    y = block(x)

    assert len(y.shape) == 4
    assert y.shape[1] == planes * block.expansion
    assert y.shape[2] == spectrogram_size // stride
    assert y.shape[3] == spectrogram_size // stride


@pytest.fixture(params=[1, 2, 3])
def in_channels(request):
    return request.param


@pytest.fixture(params=[[2, 2, 2, 2], [2, 4, 4], [3, 4, 6, 3]])
def layers(request):
    return request.param


@pytest.fixture(params=[None, "stride", "dilation"])
def stride_type(request):
    return request.param


@pytest.fixture(params=[BottleneckResNet2D, ResNet2D])
def architecture(request):
    return request.param


def test_resnet(
    architecture,
    kernel_size,
    layers,
    classes,
    in_channels,
    spectrogram_size,
    stride_type,
):
    if kernel_size % 2 == 0:
        with pytest.raises(ValueError):
            nn = ResNet2D(in_channels, layers, classes, kernel_size)
        return

    if stride_type is not None:
        stride_type = [stride_type] * (len(layers) - 1)

    if (
        stride_type is not None
        and stride_type[0] == "dilation"
        and architecture == ResNet2D
    ):
        with pytest.raises(NotImplementedError):
            nn = architecture(
                in_channels,
                layers,
                classes,
                kernel_size,
                stride_type=stride_type,
            )
        return

    nn = architecture(
        in_channels, layers, classes, kernel_size, stride_type=stride_type
    )
    x = torch.randn(8, in_channels, spectrogram_size, spectrogram_size)
    y = nn(x)
    assert y.shape == (8, classes)

    with pytest.raises(ValueError):
        stride_type = ["stride"] * len(layers)
        nn = architecture(
            in_channels, layers, kernel_size, stride_type=stride_type
        )
    with pytest.raises(ValueError):
        stride_type = ["strife"] * (len(layers) - 1)
        nn = architecture(
            in_channels, layers, kernel_size, stride_type=stride_type
        )



import h5py
import numpy as np
import pytest

from ml4gw.dataloading import Hdf5TimeSeriesDataset


class TestHdf5TimeSeriesDataset:
    @pytest.fixture
    def channels(self):
        return ["A", "B"]

    @pytest.fixture
    def sample_rate(self):
        return 128

    @pytest.fixture
    def kernel_size(self, sample_rate):
        return 2 * sample_rate

    @pytest.fixture
    def batch_size(self):
        return 128

    @pytest.fixture
    def batches_per_epoch(self):
        return 10

    @pytest.fixture
    def fnames(self, channels, sample_rate, tmp_path):
        data_dir = tmp_path / "data"
        data_dir.mkdir(exist_ok=True)

        fnames = {"a.h5": 10, "b.h5": 4, "c.h5": 6}
        fnames = {data_dir / fname: length for fname, length in fnames.items()}

        idx = 0
        keys = sorted(fnames)
        for fname in keys:
            length = fnames[fname]
            with h5py.File(fname, "w") as f:
                size = int(length * sample_rate)
                x = np.arange(idx, idx + size)
                f[channels[0]] = x
                f[channels[1]] = -x
                idx += size
        return fnames

    @pytest.fixture(params=[True, False, "files"])
    def coincident(self, request):
        return request.param

    @pytest.fixture
    def dataset(
        self,
        fnames,
        channels,
        kernel_size,
        batch_size,
        batches_per_epoch,
        coincident,
    ):
        return Hdf5TimeSeriesDataset(
            sorted(fnames.keys()),
            channels,
            kernel_size,
            batch_size,
            batches_per_epoch,
            coincident=coincident,
        )

    def test_coincident_arg_value(
        self,
        fnames,
        channels,
        kernel_size,
        batch_size,
        batches_per_epoch,
    ):
        with pytest.raises(ValueError):
            Hdf5TimeSeriesDataset(
                sorted(fnames.keys()),
                channels,
                kernel_size,
                batch_size,
                batches_per_epoch,
                coincident="wrong",
            )

    def test_init(self, dataset):
        assert dataset.num_channels == 2
        assert len(dataset) == 10
        expected_probs = np.array([0.5, 0.2, 0.3])
        np.testing.assert_equal(expected_probs, dataset.probs)

    def test_sample_fnames(self, dataset):
        fnames = dataset.sample_fnames(size=(10,))
        assert len(fnames) == 10

        # really weak check: let's at least confirm
        # that we sample the 10s segment  more than
        # we sample the 4s segment.
        counts = {fname.name: 0 for fname in dataset.fnames}
        for _ in range(10):
            fnames = dataset.sample_fnames((10,))
            for fname in fnames:
                counts[fname.name] += 1
        assert counts["a.h5"] > counts["b.h5"]

        # when not specifying fnames per batch,
        # ensure all 3 files are sampled
        fnames = dataset.sample_fnames((10000,))
        assert len(np.unique(fnames)) == 3

        # override fnames per batch for testing
        dataset.num_files_per_batch = 1
        fnames = dataset.sample_fnames(size=(10,))
        assert len(np.unique(fnames)) == 1

        # use large enough size (1000)
        # such that extremely likely
        # each file sampled at least once
        dataset.num_files_per_batch = 2
        fnames = dataset.sample_fnames(size=(1000,))
        assert len(np.unique(fnames)) == 2

    def test_sample_batch(self, dataset, kernel_size, coincident):
        x = dataset.sample_batch()
        assert x.shape == (128, 2, kernel_size)

        if coincident is True:
            # coincidence check is simple: they should all match
            for sample in x:
                assert (sample[0] == -sample[1]).all().item()
                assert (np.diff(sample[0].numpy()) == 1).all()
        else:
            # for non-coincident loading, do some checks around
            # which file each sampled channel corresponds to
            def get_bin(i):
                if i < (128 * 10):
                    return 0
                if i >= (128 * 14):
                    return 2
                return 1

            all_equal = True
            for sample in x:
                # check the files each channel came from
                bin0 = get_bin(sample[0, 0])
                bin1 = get_bin(-sample[1, 0])

                if bin0 != bin1:
                    # if they're not equal, we know that at
                    # least one batch element came out with
                    # non-coincident samples, which is good
                    # enough for us
                    all_equal = False

                    # if we're being fully non-coincident, then
                    # just the fact that we have any samples
                    # from different files means the check is done
                    if coincident is False:
                        break

                    # if we're sampling independently from within
                    # a given file, this is incorrect
                    raise ValueError(f"Got bins {bin0} and {bin1}")
                else:
                    # if the channels are from the same file,
                    # still check whether they come from the
                    # same part of the same file
                    all_equal &= (sample[0] == -sample[1]).all().item()
            else:
                # only get here if the loop never broke, which is only
                # a problem when we're sampling fully non-coincidentally
                if self.coincident is False:
                    raise ValueError("All channels came from same file")

            # if all channels came out the same, something went wrong
            if all_equal:
                raise ValueError("Sampling was coincident")

    def test_iter(self, dataset, kernel_size):
        # test_sample_batch covered most our checks, so here
        # we'll just make sure that we respect the dataset lengths
        for x in dataset:
            assert x.shape == (128, 2, kernel_size)
        assert len(dataset) == 10



from math import isclose

import numpy as np
import pytest
import torch

from ml4gw.nn.streaming import OnlineAverager
from ml4gw.utils.slicing import unfold_windows


@pytest.fixture(params=[1, 2, 4])
def batch_size(request):
    return request.param


@pytest.fixture(params=[1, 2, 4])
def update_size(request):
    return request.param


@pytest.fixture(params=[2, 10])
def num_updates(request):
    return request.param


@pytest.fixture(params=[1, 2])
def num_channels(request):
    return request.param


@pytest.fixture
def validate_output(num_updates, update_size, num_channels):
    def f(start, stop, offset, output):
        expected = np.arange(start, stop)
        for channel in output:
            for n, (i, j) in enumerate(zip(expected, channel)):
                step = (offset * update_size + n) // update_size
                factor = min((step + 1) / num_updates, 1)
                assert isclose(i * factor, j, rel_tol=1e-6)

    return f


def test_online_averager(
    batch_size, update_size, num_updates, num_channels, validate_output
):
    averager = OnlineAverager(
        update_size=update_size,
        batch_size=batch_size,
        num_updates=num_updates,
        num_channels=num_channels,
    )

    # make a batch of overlapping aranged data such that
    # the online average is just the values themselves
    kernel_size = 100
    total_updates = 2 * batch_size - 1
    size = total_updates * update_size + kernel_size
    x = torch.arange(size)[None].type(torch.float32)
    x = unfold_windows(x, kernel_size, update_size)
    x = torch.repeat_interleave(x, num_channels, axis=1)
    assert x.size(0) == 2 * batch_size

    # initialize a blank initial snapshot
    state = averager.get_initial_state()

    # perform the first aggregation step
    stream, new_state = averager(x[:batch_size], state)

    # make sure the shapes are right
    expected_shape = (num_channels, update_size * batch_size)
    assert stream.shape == expected_shape
    assert new_state.shape == state.shape

    # now validate that the streamed value is correct
    start = kernel_size - update_size * num_updates
    stop = start + update_size * batch_size
    validate_output(start, stop, 0, stream.cpu().numpy())

    # now take the next step and confirm everything again
    stream, newer_state = averager(x[batch_size:], new_state)

    assert stream.shape == expected_shape
    assert newer_state.shape == state.shape

    start += update_size * batch_size
    stop = start + update_size * batch_size
    validate_output(start, stop, batch_size, stream.cpu().numpy())



import astropy.units as u
import numpy as np
import pytest
import torch
from lalsimulation.gwsignal.core.waveform import (
    GenerateTDWaveform,
    LALCompactBinaryCoalescenceGenerator,
)

from ml4gw.waveforms import IMRPhenomD, conversion
from ml4gw.waveforms.generator import TimeDomainCBCWaveformGenerator


@pytest.fixture()
def num_samples():
    return 1000


@pytest.fixture(params=[2048])
def sample_rate(request):
    return request.param


class GWsignalGenerator(LALCompactBinaryCoalescenceGenerator):
    """
    Override gwsignal class to enforce use of `gwsignal` conditioning routines,
    and to enforce that the frequency domain version of the approximant is used
    for generating the waveform.
    """

    @property
    def metadata(self):
        metadata = {
            "type": "cbc_lalsimulation",
            "f_ref_spin": True,
            "modes": True,
            "polarizations": True,
            "implemented_domain": self._implemented_domain,
            "generation_domain": self._generation_domain,
            "approximant": self._approx_name,
            "implementation": "LALSimulation",
            "conditioning_routines": "gwsignal",
        }
        return metadata

    def _update_domains(self):
        self._implemented_domain = "freq"


def test_cbc_waveform_generator(
    chirp_mass,
    mass_ratio,
    chi1,
    chi2,
    phase,
    distance_far,
    theta_jn,
    sample_rate,
):
    # torch.manual_seed(10)
    duration = 10
    f_min = 20
    f_ref = 40
    right_pad = 0.5

    generator = TimeDomainCBCWaveformGenerator(
        approximant=IMRPhenomD(),
        sample_rate=sample_rate,
        duration=duration,
        f_min=f_min,
        f_ref=f_ref,
        right_pad=right_pad,
    )

    mass_1, mass_2 = conversion.chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )
    s1x = torch.zeros_like(chi1)
    s1y = torch.zeros_like(chi1)
    s1z = chi1
    s2x = torch.zeros_like(chi2)
    s2y = torch.zeros_like(chi2)
    s2z = chi2
    ml4gw_parameters = {
        "chirp_mass": chirp_mass,
        "mass_ratio": mass_ratio,
        "mass_1": mass_1,
        "mass_2": mass_2,
        "chi1": chi1,
        "chi2": chi2,
        "s1z": s1z,
        "s2z": s2z,
        "s1x": s1x,
        "s1y": s1y,
        "s2x": s2x,
        "s2y": s2y,
        "phic": phase,
        "distance": distance_far,
        "inclination": theta_jn,
    }
    hc_ml4gw, hp_ml4gw = generator(**ml4gw_parameters)

    # now compare each waveform with lalsimulation SimInspiralTD
    for i in range(len(chirp_mass)):
        # construct gwsignal params
        gwsignal_params = {
            "mass1": ml4gw_parameters["mass_1"][i].item() * u.solMass,
            "mass2": ml4gw_parameters["mass_2"][i].item() * u.solMass,
            "deltaT": (1 / sample_rate) * u.s,
            "f22_start": f_min * u.Hz,
            "f22_ref": f_ref * u.Hz,
            "phi_ref": ml4gw_parameters["phic"][i].item() * u.rad,
            "distance": (ml4gw_parameters["distance"][i].item() * u.Mpc),
            "inclination": ml4gw_parameters["inclination"][i].item() * u.rad,
            "eccentricity": 0.0 * u.dimensionless_unscaled,
            "longAscNodes": 0.0 * u.rad,
            "meanPerAno": 0.0 * u.rad,
            "spin1x": ml4gw_parameters["s1x"][i].item()
            * u.dimensionless_unscaled,
            "spin1y": ml4gw_parameters["s1y"][i].item()
            * u.dimensionless_unscaled,
            "spin1z": ml4gw_parameters["s1z"][i].item()
            * u.dimensionless_unscaled,
            "spin2x": ml4gw_parameters["s2x"][i].item()
            * u.dimensionless_unscaled,
            "spin2y": ml4gw_parameters["s2y"][i].item()
            * u.dimensionless_unscaled,
            "spin2z": ml4gw_parameters["s2z"][i].item()
            * u.dimensionless_unscaled,
            "condition": 1,
        }

        gwsignal_generator = GWsignalGenerator("IMRPhenomD")
        hp_gwsignal, hc_gwsignal = GenerateTDWaveform(
            gwsignal_params, gwsignal_generator
        )

        hp = hp_ml4gw[i].detach().numpy()
        hc = hc_ml4gw[i].detach().numpy()

        # now align the gwsignal and ml4gw waveforms so we can compoare
        ml4gw_times = np.arange(0, len(hp) / sample_rate, 1 / sample_rate)
        ml4gw_times -= duration - right_pad

        # TODO: track this down

        # gwsignal will adjust the "epoch"
        # i.e. the coalescence time based on
        # the maximum value of the array.
        # In some cases, the ml4gw prediction
        # for the time of maximum value is slightly different
        # than the gwsignal/lalsimulation. Typically, either
        # the maximum, or second maximum will align with gwsignal.
        # Presumably this is due to accumulation of discrepancies
        # between our implementation and gwsignal/lalsimulations.

        # So, for testing we align the waveform using both
        # the maximum and second maximum values, and assert True
        # if any of them satisfy the tolerances

        for ml4gw_pol, gwsignal_pol in zip(
            [hp, hc], [hp_gwsignal, hc_gwsignal]
        ):
            argsorted = np.argsort(ml4gw_pol)[::-1]
            for j in range(2):
                argmax = argsorted[j]
                ml4gw_times = ml4gw_times - ml4gw_times[argmax]

                max_time = min(ml4gw_times[-1], gwsignal_pol.times.value[-1])
                min_time = max(ml4gw_times[0], gwsignal_pol.times.value[0])

                mask = gwsignal_pol.times.value <= max_time
                mask &= gwsignal_pol.times.value >= min_time

                ml4gw_mask = ml4gw_times <= max_time
                ml4gw_mask &= ml4gw_times >= min_time

                close = np.allclose(
                    gwsignal_pol.value[mask],
                    ml4gw_pol[ml4gw_mask],
                    atol=4e-23,
                    rtol=0.01,
                )

                if close:
                    break

                # TODO: track this down

                # theres an off by one error that occurs
                # occasionally when attempting to align the
                # gwsignal and ml4gw waveforms that is causing
                # testing comparison issues, so assert that
                # either one of them is close enough

                for k in range(1, 2):
                    if close:
                        break

                    close_shifted = np.allclose(
                        gwsignal_pol.value[mask][:-k],
                        ml4gw_pol[ml4gw_mask][k:],
                        atol=4e-23,
                        rtol=0.01,
                    )

                    close_shifted = close_shifted or np.allclose(
                        gwsignal_pol.value[mask][k:],
                        ml4gw_pol[ml4gw_mask][:-k],
                        atol=4e-23,
                        rtol=0.01,
                    )

                    close = close or close_shifted

            assert close



import lalsimulation
import pytest
import torch

from ml4gw.constants import MSUN
from ml4gw.waveforms.cbc import utils

NUM_SAMPLES = 100


@pytest.fixture(params=[torch.rand(NUM_SAMPLES) + 1 * 15])
def fstart(request):
    return torch.tensor(request.param).double()


@pytest.fixture
def masses():
    m1 = (
        torch.rand(NUM_SAMPLES, dtype=torch.float64) * 45 + 5
    )  # Uniform distribution between 5 and 50
    m2 = (
        torch.rand(NUM_SAMPLES, dtype=torch.float64) * 45 + 5
    )  # Uniform distribution between 5 and 50
    return m1 * MSUN, m2 * MSUN


@pytest.fixture
def spins():
    s1 = (
        torch.rand(NUM_SAMPLES) * 2 - 1
    )  # Uniform distribution between -1 and 1
    s2 = (
        torch.rand(NUM_SAMPLES) * 2 - 1
    )  # Uniform distribution between -1 and 1
    return s1.double(), s2.double()


def test_chirp_time_bound(masses, spins, fstart):
    m1, m2 = masses
    s1, s2 = spins

    result = utils.chirp_time_bound(fstart, m1, m2, s1, s2)
    for i, (fs, m1_val, m2_val, s1_val, s2_val) in enumerate(
        zip(fstart, m1, m2, s1, s2)
    ):
        expected = lalsimulation.SimInspiralChirpTimeBound(
            fs.item(),
            m1_val.item(),
            m2_val.item(),
            s1_val.item(),
            s2_val.item(),
        )
        assert torch.isclose(
            result[i].float(), torch.tensor(expected), rtol=1e-3
        )


def test_final_black_hole_spin_bound(spins):
    s1, s2 = spins

    result = utils.final_black_hole_spin_bound(s1, s2)

    for i, (s1_val, s2_val) in enumerate(zip(s1, s2)):
        expected = lalsimulation.SimInspiralFinalBlackHoleSpinBound(
            s1_val.item(), s2_val.item()
        )
        assert torch.isclose(
            result[i].float(), torch.tensor(expected), rtol=1e-3
        )


def test_merge_time_bound(masses):
    m1, m2 = masses

    result = utils.merge_time_bound(m1, m2)

    for i, (m1_val, m2_val) in enumerate(zip(m1, m2)):
        expected = lalsimulation.SimInspiralMergeTimeBound(
            m1_val.item(), m2_val.item()
        )
        assert torch.isclose(
            result[i].float(), torch.tensor(expected), rtol=1e-3
        )


def test_ringdown_time_bound(masses, spins):
    m1, m2 = masses
    s1, s2 = spins
    s = utils.final_black_hole_spin_bound(s1, s2)
    total_mass = m1 + m2

    result = utils.ringdown_time_bound(total_mass, s)

    for i, (total_mass_val, s_val) in enumerate(zip(total_mass, s)):
        expected = lalsimulation.SimInspiralRingdownTimeBound(
            total_mass_val.item(), s_val.item()
        )
        assert torch.isclose(
            result[i].float(), torch.tensor(expected), rtol=1e-3
        )


def test_chirp_start_frequency_bound(fstart, spins, masses):
    m1, m2 = masses
    s1, s2 = spins
    tchirp = utils.chirp_time_bound(fstart, m1, m2, s1, s2)
    result = utils.chirp_start_frequency_bound(tchirp, m1, m2)

    for i, (tchirp_val, m1_val, m2_val) in enumerate(zip(tchirp, m1, m2)):
        expected = lalsimulation.SimInspiralChirpStartFrequencyBound(
            tchirp_val.item(), m1_val.item(), m2_val.item()
        )
        assert torch.isclose(
            result[i].float(), torch.tensor(expected), rtol=1e-3
        )



import numpy as np
import pytest
import torch
from lalinference import BurstSineGaussian

from ml4gw.waveforms import SineGaussian


@pytest.fixture(params=[2048, 4096])
def sample_rate(request):
    return request.param


@pytest.fixture(params=[2.0, 4.0, 8.0])
def duration(request):
    return request.param


@pytest.fixture(params=[3.0, 10.0, 100.0, 55.0])
def quality(request):
    return torch.tensor(request.param, dtype=torch.float64)


@pytest.fixture(params=[100.0, 500.0, 800.0, 961.0])
def frequency(request):
    return torch.tensor(request.param, dtype=torch.float64)


# for amplitudes above ~7e-20, the difference between torch imp
# and lalsim is > ~1e-24. Our implementations are 1 to 1, so
# discrep must be from numerical issues?
@pytest.fixture(params=[1e-23, 1e-22, 1e-21, 1e-20, 7e-20])
def hrss(request):
    return torch.tensor(request.param, dtype=torch.float64)


@pytest.fixture(params=[0.0, np.pi / 2.0, np.pi, 2 * np.pi])
def phase(request):
    return torch.tensor(request.param, dtype=torch.float64)


@pytest.fixture(
    params=[
        0.0,
        0.5,
        1.0,
        0.1,
    ]
)
def eccentricity(request):
    return torch.tensor(request.param, dtype=torch.float64)


def test_sine_gaussian(
    duration,
    sample_rate,
    quality,
    frequency,
    hrss,
    phase,
    eccentricity,
):
    sine_gaussian = SineGaussian(sample_rate, duration)

    # calculate waveforms with torch implementation
    cross, plus = sine_gaussian(quality, frequency, hrss, phase, eccentricity)
    cross, plus = cross[0].numpy(), plus[0].numpy()

    quality = quality.item()
    frequency = frequency.item()
    phase = phase.item()
    eccentricity = eccentricity.item()
    hrss = hrss.item()

    # calculate waveform with lalsimulation
    hplus, hcross = BurstSineGaussian(
        Q=quality,
        centre_frequency=frequency,
        hrss=hrss,
        eccentricity=eccentricity,
        phase=phase,
        delta_t=1 / sample_rate,
    )
    hplus = hplus.data.data
    hcross = hcross.data.data

    # compare cross and plus polarizations
    n_samples = len(hplus)
    start, stop = (
        len(cross) // 2 - n_samples // 2,
        len(cross) // 2 + n_samples // 2 + 1,
    )
    cross, plus = cross[start:stop], plus[start:stop]

    assert np.allclose(
        cross, hcross, atol=1e-24
    )  # factor of 10 smaller than ligo noise floor
    assert np.allclose(
        plus, hplus, atol=1e-24
    )  # factor of 10 smaller than ligo noise floor



import numpy as np
import pytest
import torch

from ml4gw.nn.streaming import Snapshotter


@pytest.fixture(params=[1, 4, 100])
def snapshot_size(request):
    return request.param


@pytest.fixture(params=[1, 2, 4])
def batch_size(request):
    return request.param


@pytest.fixture(params=[1, 2, 4])
def stride_size(request):
    return request.param


@pytest.fixture(params=[[1], 1, [4], 4, [1, 4], [1, 2, 4]])
def channels(request):
    return request.param


@pytest.fixture
def snapshotter(snapshot_size, stride_size, batch_size, channels):
    return Snapshotter(snapshot_size, stride_size, batch_size, channels)


def test_snapshotter(snapshot_size, stride_size, batch_size, channels):
    if isinstance(channels, int):
        num_channels = channels
        channels = [num_channels]
        channels_per_snapshot = None
    else:
        num_channels = sum(channels)
        channels_per_snapshot = channels

    # test that we don't allow for snapshotting
    # when the stride is longer than the kernel,
    # since there's no real stateful behavior
    # in this case
    if stride_size >= snapshot_size:
        with pytest.raises(ValueError):
            snapshotter = Snapshotter(
                num_channels,
                snapshot_size,
                stride_size,
                batch_size,
                channels_per_snapshot,
            )
        return

    # now make sure we test for agreement between
    # num_channels and any way of breaking up the channels
    if channels_per_snapshot is not None:
        with pytest.raises(ValueError):
            snapshotter = Snapshotter(
                num_channels + 1,
                snapshot_size,
                stride_size,
                batch_size,
                channels_per_snapshot,
            )

    snapshotter = Snapshotter(
        num_channels,
        snapshot_size,
        stride_size,
        batch_size,
        channels_per_snapshot,
    )

    # now run an input through as a new sequence and
    # make sure we get the appropriate number of outputs
    offset = torch.arange(num_channels)[:, None] * snapshot_size

    update_size = stride_size * batch_size
    snapshot = snapshotter.get_initial_state() + 1
    snapshot = torch.cumsum(snapshot, axis=-1) - 1
    snapshot += offset

    update = torch.arange(update_size) + 1 + snapshot[0, -1]
    update = update.view(1, -1).repeat(num_channels, 1)
    update += offset

    outputs = snapshotter(update, snapshot)
    outputs = [i.cpu().numpy() for i in outputs]
    new_snapshot = outputs.pop(-1)

    # make sure we have as many outputs as we
    # specified number of channel groups
    assert len(outputs) == len(channels)

    # now go through and make sure that each
    # of the windows in each batch contains
    # the expected data
    offset = 0
    for output, channel_dim in zip(outputs, channels):
        expected = (batch_size, channel_dim, snapshot_size)
        assert output.shape == tuple(expected)

        for i, row in enumerate(output):
            for j, channel in enumerate(row):
                start = j * snapshot_size + i * stride_size
                stop = start + snapshot_size
                expected = np.arange(start, stop) + offset
                np.testing.assert_equal(channel, expected)
        offset += channel_dim * snapshot_size

    expected_size = snapshot_size - stride_size
    assert new_snapshot.shape == (num_channels, expected_size)
    for i, channel in enumerate(new_snapshot):
        start = update_size + i * snapshot_size
        stop = start + expected_size
        expected = np.arange(start, stop)
        np.testing.assert_equal(channel, expected)



import numpy as np
import pytest
import torch
from scipy.interpolate import RectBivariateSpline, UnivariateSpline
from torch import Tensor

from ml4gw.transforms import SplineInterpolate


class TestSplineInterpolate:
    @pytest.fixture(params=[50, 100, 200])
    def x_out_len(self, request):
        return request.param

    @pytest.fixture(params=[25, 200, 1000])
    def y_out_len(self, request):
        return request.param

    def test_1d_interpolation(self, x_out_len):
        x_in = np.linspace(0, 10, 100)
        data = np.sin(x_in)
        # There are edge effects in the torch transform that
        # aren't present in scipy. Would be great to solve that,
        # but a workaround is to interpolate well within the
        # boundaries of the input coordinates. Unfortunately,
        # what specifically that means depends on the size of
        # the input array.
        pad = len(x_in) // 10
        x_out = np.linspace(x_in[pad], x_in[-pad], x_out_len)

        scipy_spline = UnivariateSpline(x_in, data, k=3, s=0)
        expected = scipy_spline(x_out)

        torch_spline = SplineInterpolate(
            x_in=Tensor(x_in),
            x_out=Tensor(x_out),
            kx=3,
        )
        actual = torch_spline(Tensor(data)).squeeze().numpy()

        # The "steady-state" ratio between the torch and scipy
        # interpolations is about 0.9990, with some minor fluctuations.
        # Would be nice to know why the torch interpolation is
        # consistently smaller
        assert np.allclose(actual, expected, rtol=1e-2)

    def test_2d_interpolation(self, x_out_len, y_out_len):
        x_in = np.linspace(0, 10, 100)
        y_in = np.linspace(0, 5, 200)
        x_grid, y_grid = np.meshgrid(x_in, y_in)
        data = np.sin(x_grid) * np.cos(y_grid)

        pad = len(x_in) // 10
        x_out = np.linspace(x_in[pad], x_in[-pad], x_out_len)
        pad = len(y_in) // 10
        y_out = np.linspace(y_in[pad], y_in[-pad], y_out_len)

        scipy_spline = RectBivariateSpline(x_in, y_in, data.T, kx=3, ky=3, s=0)
        expected = scipy_spline(x_out, y_out).T

        torch_spline = SplineInterpolate(
            x_in=Tensor(x_in),
            x_out=Tensor(x_out),
            y_in=Tensor(y_in),
            y_out=Tensor(y_out),
            kx=3,
            ky=3,
        )
        actual = torch_spline(Tensor(data)).squeeze().numpy()

        # The "steady-state" ratio between the torch and scipy
        # interpolations is about 0.999, with some minor fluctuations.
        # Would be nice to know why the torch interpolation is
        # consistently smaller
        assert np.allclose(actual, expected, rtol=5e-3)

    def test_errors(self):
        x_in = torch.arange(10)
        x_out = x_in
        torch_spline = SplineInterpolate(x_in)
        data = torch.randn(len(x_in))
        with pytest.raises(ValueError) as exc:
            torch_spline(data)
        assert str(exc.value).startswith("Output x-coordinates were not")

        data = torch.randn((1, 2, 3, 4, 5))
        with pytest.raises(ValueError) as exc:
            torch_spline(data, x_out=x_out)
        assert str(exc.value).startswith("Input data has more than 4")

        y_in = torch.arange(10)
        torch_spline = SplineInterpolate(x_in=x_in, y_in=y_in)
        data = torch.randn(len(x_in))
        with pytest.raises(ValueError) as exc:
            torch_spline(data, x_out=x_out)
        assert str(exc.value).startswith("An input y-coordinate array")

        data = torch.randn((len(y_in) - 1, len(x_in) - 1))
        with pytest.raises(ValueError) as exc:
            torch_spline(data, x_out=x_out)
        assert str(exc.value).startswith("The spatial dimensions of the data")



import pytest
import torch

from ml4gw.transforms import ShiftedPearsonCorrelation


class TestShiftedPearsonCorrelation:
    @pytest.fixture
    def max_shift(self):
        return 5

    @pytest.fixture
    def transform(self, max_shift):
        return ShiftedPearsonCorrelation(max_shift)

    def check_in_range(self, corr):
        in_range = (-1 <= corr) & (corr <= 1)
        assert in_range.all().item()

    def test_x_ndim_3(self, transform, max_shift):
        expected_shape = (2 * max_shift + 1, 4, 2)

        x = torch.randn(2, 2048)

        # set up a y which is just a shifted
        # version of x at each batch index
        y = torch.zeros((4, 2, 2048))
        for i in range(4):
            j = i - 2
            if j < 0:
                y[i, :, -j:] = x[:, :j]
            elif j > 0:
                y[i, :, :-j] = x[:, j:]
            else:
                y[i] = x

        # make all batch elements of x the same
        x = x.view(1, 2, -1).repeat(4, 1, 1)
        corr = transform(x, y)

        # first check that we have the expected shape
        assert corr.shape == expected_shape

        # check that all our values are in the expected range
        self.check_in_range(corr)

        # check that we get our maximum matches
        # at the expected indices
        idx = corr[:, :, 0].argmax(dim=0)
        expected_shifts = torch.arange(4) + max_shift - 2
        assert torch.equal(idx, expected_shifts)

        # and that those matches are nearly 1
        maxs = corr[idx, torch.arange(4), 0]
        expected_max = torch.ones(4)
        assert torch.allclose(maxs, expected_max, rtol=0.01)

        # do similar checks for 2dim y
        corr = transform(x, y[0])
        assert corr.shape == expected_shape
        self.check_in_range(corr)

        idx = corr[:, :, 0].argmax(dim=0)
        expected_shifts = torch.ones(4) * (max_shift - 2)
        assert torch.equal(idx, expected_shifts)

        maxs = corr[max_shift - 2, :, 0]
        assert torch.allclose(maxs, expected_max, rtol=0.01)

        # and finally for 1dim y. We've only been checking
        # against y's first channel, so this will just
        # look exactly like the last one.
        corr = transform(x, y[0, 0])
        assert corr.shape == expected_shape
        self.check_in_range(corr)

        idx = corr[:, :, 0].argmax(dim=0)
        assert torch.equal(idx, expected_shifts)

        maxs = corr[max_shift - 2, :, 0]
        assert torch.allclose(maxs, expected_max, rtol=0.01)



import lal
import lalsimulation
import numpy as np
import pytest
import torch
from astropy import units as u
from scipy.signal import filtfilt, iirfilter

from ml4gw.constants import MSUN
from ml4gw.transforms.iirfilter import IIRFilter
from ml4gw.waveforms.conversion import (
    bilby_spins_to_lalsim,
    chirp_mass_and_mass_ratio_to_components,
)


@pytest.fixture
def low_cutoff():
    return 100


@pytest.fixture
def high_cutoff():
    return 100


@pytest.fixture(
    params=[
        (
            0.5,
            None,
            "cheby1",
        ),
        (None, 20, "cheby2"),
        (
            0.5,
            20,
            "ellip",
        ),
        (
            None,
            None,
            "bessel",
        ),
        (None, None, "butter"),
    ]
)
def rp_rs_filter(request):
    return request.param


@pytest.fixture(params=[256, 512, 1024, 2048])
def sample_rate(request):
    return request.param


@pytest.fixture(params=[2, 4, 6, 8])
def order(request):
    return request.param


def test_filters_synthetic_signal(
    sample_rate, order, low_cutoff, high_cutoff, rp_rs_filter
):
    t = np.linspace(0, 1.0, sample_rate, endpoint=False)
    tone_freq = 50
    noise_amplitude = 0.5

    signal = np.sin(2 * np.pi * tone_freq * t)
    noise = noise_amplitude * np.random.normal(size=t.shape)
    combined_signal = signal + noise

    slice_length = int(0.15 * sample_rate)

    rp, rs, filter = rp_rs_filter

    b, a = iirfilter(
        order,
        low_cutoff,
        btype="low",
        analog=False,
        output="ba",
        fs=sample_rate,
        rp=rp,
        rs=rs,
        ftype=filter,
    )
    scipy_filtered_data_low = filtfilt(b, a, combined_signal)[
        slice_length:-slice_length
    ]

    b, a = iirfilter(
        order,
        high_cutoff,
        btype="high",
        analog=False,
        output="ba",
        fs=sample_rate,
        rp=rp,
        rs=rs,
        ftype=filter,
    )
    scipy_filtered_data_high = filtfilt(b, a, combined_signal)[
        slice_length:-slice_length
    ]

    # test one of these with a tensor input instead of scalar Wn, rs, rps
    torch_filtered_data_low = IIRFilter(
        order,
        torch.tensor(low_cutoff),
        btype="low",
        analog=False,
        fs=sample_rate,
        rs=torch.tensor(rs) if rs is not None else None,
        rp=torch.tensor(rp) if rp is not None else None,
        ftype=filter,
    )(torch.tensor(combined_signal).repeat(10, 1))[
        :, slice_length:-slice_length
    ].numpy()

    torch_filtered_data_high = IIRFilter(
        order,
        high_cutoff,
        btype="high",
        analog=False,
        fs=sample_rate,
        rs=rs,
        rp=rp,
        ftype=filter,
    )(torch.tensor(combined_signal).repeat(10, 1))[
        :, slice_length:-slice_length
    ].numpy()

    # test batch processing
    for i in range(9):
        assert np.allclose(
            torch_filtered_data_low[0],
            torch_filtered_data_low[i + 1],
            atol=float(np.finfo(float).eps),
        )
        assert np.allclose(
            torch_filtered_data_high[0],
            torch_filtered_data_high[i + 1],
            atol=float(np.finfo(float).eps),
        )

    assert np.allclose(
        scipy_filtered_data_low,
        torch_filtered_data_low[0],
        atol=2e-1,
    )
    assert np.allclose(
        scipy_filtered_data_high,
        torch_filtered_data_high[0],
        atol=2e-1,
    )


@pytest.fixture()
def num_samples():
    return 1


@pytest.fixture(params=[20, 40])
def f_ref(request):
    return request.param


def test_filters_phenom_signal(
    sample_rate,
    order,
    rp_rs_filter,
    low_cutoff,
    high_cutoff,
    chirp_mass,
    mass_ratio,
    distance,
    phase,
    f_ref,
    theta_jn,
    phi_jl,
    tilt_1,
    tilt_2,
    phi_12,
    a_1,
    a_2,
):
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )
    (
        inclination,
        chi1x,
        chi1y,
        chi1z,
        chi2x,
        chi2y,
        chi2z,
    ) = bilby_spins_to_lalsim(
        theta_jn,
        phi_jl,
        tilt_1,
        tilt_2,
        phi_12,
        a_1,
        a_2,
        mass_1,
        mass_2,
        f_ref,
        phase,
    )

    params = {
        "m1": mass_1.item() * MSUN,
        "m2": mass_2.item() * MSUN,
        "S1x": chi1x.item(),
        "S1y": chi1y.item(),
        "S1z": chi1z.item(),
        "S2x": chi2x.item(),
        "S2y": chi2y.item(),
        "S2z": chi2z.item(),
        "distance": (distance.item() * u.Mpc).to("m").value,
        "inclination": inclination.item(),
        "phiRef": phase.item(),
        "longAscNodes": 0.0,
        "eccentricity": 0.0,
        "meanPerAno": 0.0,
        "deltaF": 1.0 / sample_rate,
        "f_min": 10.0,
        "f_ref": f_ref,
        "f_max": 300,
        "approximant": lalsimulation.IMRPhenomPv2,
        "LALpars": lal.CreateDict(),
    }
    hp_lal, _ = lalsimulation.SimInspiralChooseFDWaveform(**params)
    hp_lal = hp_lal.data.data.real

    slice_length = int(0.15 * sample_rate)

    rp, rs, filter = rp_rs_filter

    b, a = iirfilter(
        order,
        low_cutoff,
        btype="low",
        analog=False,
        output="ba",
        fs=sample_rate,
        rp=rp,
        rs=rs,
        ftype=filter,
    )

    scipy_filtered_data_low = filtfilt(b, a, hp_lal)[
        slice_length:-slice_length
    ]

    b, a = iirfilter(
        order,
        high_cutoff,
        btype="high",
        analog=False,
        output="ba",
        fs=sample_rate,
        rp=rp,
        rs=rs,
        ftype=filter,
    )
    scipy_filtered_data_high = filtfilt(b, a, hp_lal)[
        slice_length:-slice_length
    ]

    torch_filtered_data_low = IIRFilter(
        order,
        low_cutoff,
        btype="low",
        analog=False,
        fs=sample_rate,
        rs=rs,
        rp=rp,
        ftype=filter,
    )(torch.tensor(hp_lal).repeat(10, 1))[
        :, slice_length:-slice_length
    ].numpy()

    torch_filtered_data_high = IIRFilter(
        order,
        high_cutoff,
        btype="high",
        analog=False,
        fs=sample_rate,
        rs=rs,
        rp=rp,
        ftype=filter,
    )(torch.tensor(hp_lal).repeat(10, 1))[
        :, slice_length:-slice_length
    ].numpy()

    # test batch processing
    for i in range(9):
        assert np.allclose(
            torch_filtered_data_low[0],
            torch_filtered_data_low[i + 1],
            atol=float(np.finfo(float).eps),
        )
        assert np.allclose(
            torch_filtered_data_high[0],
            torch_filtered_data_high[i + 1],
            atol=float(np.finfo(float).eps),
        )

    assert np.allclose(
        1e21 * scipy_filtered_data_low,
        1e21 * torch_filtered_data_low[0],
        atol=7e-3,
    )
    assert np.allclose(
        1e21 * scipy_filtered_data_high,
        1e21 * torch_filtered_data_high[0],
        atol=7e-3,
    )
